from enum import Enum

import ufl
from dolfinx import fem

from festim import k_B
from festim.helpers import as_fenics_constant


class SolubilityLaw(Enum):
    SIEVERT = 10
    HENRY = 20
    NONE = 30

    @classmethod
    def from_string(cls, s: str):
        """Can be removed with Python 3.11+."""
        s = s.lower()
        if s == "sievert":
            return cls.SIEVERT
        elif s == "henry":
            return cls.HENRY
        elif s == "none":
            return cls.NONE
        else:
            raise ValueError(
                "solubility_law must be one of 'sievert', 'henry', or 'none'"
            )


class Material:
    """
    Material class

    Args:
        D_0: the pre-exponential factor of the
            diffusion coefficient (m2/s)
        E_D: the activation energy of the diffusion
            coeficient (eV)
        K_S_0: the pre-exponential factor of the
            solubility coefficient (H/m3/Pa0.5)
        E_K_S: the activation energy of the solubility
            coeficient (eV)
        name: the name of the material
        thermal_conductivity: the thermal conductivity of the material (W/m/K)
        density: the density of the material (kg/m3)
        heat_capacity: the heat capacity of the material (J/kg/K)
        solubility_law: the solubility law of the material (SIEVERT, HENRY or NONE).
            For single material problems one can use NONE. This does not work for
            multi-material problems

        D: the diffusion coefficient of the material (m2/s)

    Attributes:
        D_0: the pre-exponential factor of the
            diffusion coefficient (m2/s)
        E_D: the activation energy of the diffusion
            coeficient (eV)
        K_S_0: the pre-exponential factor of the
            solubility coefficient (H/m3/Pa0.5)
        E_K_S: the activation energy of the solubility
            coeficient (eV)
        name: the name of the material
        thermal_conductivity: the thermal conductivity of the material (W/m/K)
        density: the density of the material (kg/m3)
        heat_capacity: the heat capacity of the material (J/kg/K)
        solubility_law: the solubility law of the material (SIEVERT, HENRY or NONE).
            For single material problems one can use NONE. This does not work for
            multi-material problems

    Examples:
        .. testsetup:: Material

            from festim import Material

        .. testcode:: Material

            # if only one species:
            Material(D_0=1.9e-7, E_D=0.2, name="my_mat")

            # if several species:
            Material(
                D_0={"Species_1": 1.9e-7, "Species_2": 2.0e-7},
                E_D={"Species_1": 0.2, "Species_2": 0.3},
                name="my_mat"
            )
    """

    def __init__(
        self,
        D_0: float | int | fem.Function | dict[float, int] | None = None,
        E_D: float | int | fem.Function | dict[float, int] | None = None,
        K_S_0: float | int | dict[float, int] | None = None,
        E_K_S: float | int | dict[float, int] | None = None,
        thermal_conductivity: float | None = None,
        density: float | None = None,
        heat_capacity: float | None = None,
        name: str | None = None,
        solubility_law: SolubilityLaw | str = SolubilityLaw.NONE,
        D: fem.Function | None = None,
    ) -> None:
        self.D_0 = D_0
        self.E_D = E_D
        self.K_S_0 = K_S_0
        self.E_K_S = E_K_S

        self.thermal_conductivity = thermal_conductivity
        self.density = density
        self.heat_capacity = heat_capacity
        self.name = name
        self.solubility_law = solubility_law
        self.D = D

        if self.D_0 and self.D:
            raise ValueError(
                "D_0 and D cannot be set at the same time. Please set only one of them."
            )

    @property
    def solubility_law(self):
        return self._solubility_law

    @solubility_law.setter
    def solubility_law(self, value):
        if isinstance(value, SolubilityLaw):
            self._solubility_law = value
        elif isinstance(value, str):
            self._solubility_law = SolubilityLaw.from_string(value)
        else:
            raise ValueError(f"Could not set solubility law {value=}")

    @property
    def D(self):
        return self._D

    @D.setter
    def D(self, value):
        if value is None:
            self._D = None
        elif isinstance(value, fem.Function):
            self._D = value
        else:
            raise TypeError("D must be of type fem.Function")

    def get_D_0(self, species=None):
        """Returns the pre-exponential factor of the diffusion coefficient

        Args:
            species (festim.Species or str, optional): the species we want the
                pre-exponential factor of the diffusion coefficient of. Only needed if
                D_0 is a dict.

        Returns:
            float: the pre-exponential factor of the diffusion coefficient
        """

        if isinstance(self.D_0, float | int):
            return self.D_0

        elif isinstance(self.D_0, dict):
            if species is None:
                raise ValueError("species must be provided if D_0 is a dict")

            if species in self.D_0:
                return self.D_0[species]
            elif species.name in self.D_0:
                return self.D_0[species.name]
            else:
                raise ValueError(f"{species} is not in D_0 keys")

        else:
            raise TypeError("D_0 must be either a float, int or a dict")

    def get_E_D(self, species=None):
        """Returns the activation energy of the diffusion coefficient

        Args:
            species (festim.Species or str, optional): the species we want the
                activation energy of the diffusion coefficient of. Only needed if E_D is
                a dict.

        Returns:
            float: the activation energy of the diffusion coefficient
        """

        if isinstance(self.E_D, float | int):
            return self.E_D

        elif isinstance(self.E_D, dict):
            if species is None:
                raise ValueError("species must be provided if E_D is a dict")

            if species in self.E_D:
                return self.E_D[species]
            elif species.name in self.E_D:
                return self.E_D[species.name]
            else:
                raise ValueError(f"{species} is not in E_D keys")

        else:
            raise TypeError("E_D must be either a float, int or a dict")

    def get_K_S_0(self, species=None) -> float:
        """Returns the pre-exponential factor of the solubility coefficient

        Args:
            species: the species we want the pre-exponential
                factor of the solubility coefficient of. Only needed if K_S_0 is a dict.

        Returns:
            the pre-exponential factor of the solubility coefficient
        """

        if isinstance(self.K_S_0, float | int):
            return self.K_S_0

        elif isinstance(self.K_S_0, dict):
            if species is None:
                raise ValueError("species must be provided if K_S_0 is a dict")

            if species in self.K_S_0:
                return self.K_S_0[species]
            elif species.name in self.K_S_0:
                return self.K_S_0[species.name]
            else:
                raise ValueError(f"{species} is not in K_S_0 keys")

        else:
            raise TypeError("K_S_0 must be either a float, int or a dict")

    def get_E_K_S(self, species=None) -> float:
        """Returns the activation energy of the solubility coefficient

        Args:
            species: the species we want the activation
                energy of the solubility coefficient of. Only needed if E_K_S is a dict.

        Returns:
            the activation energy of the solubility coefficient
        """

        if isinstance(self.E_K_S, float | int):
            return self.E_K_S

        elif isinstance(self.E_K_S, dict):
            if species is None:
                raise ValueError("species must be provided if E_K_S is a dict")

            if species in self.E_K_S:
                return self.E_K_S[species]
            elif species.name in self.E_K_S:
                return self.E_K_S[species.name]
            else:
                raise ValueError(f"{species} is not in E_K_S keys")

        else:
            raise TypeError("E_K_S must be either a float, int or a dict")

    def get_diffusion_coefficient(self, mesh=None, temperature=None, species=None):
        """Defines the diffusion coefficient

        Args:

            mesh (dolfinx.mesh.Mesh): the domain mesh
            temperature (dolfinx.fem.Constant): the temperature
            species (festim.Species, optional): the species we want the diffusion
                coefficient of. Only needed if D_0 and E_D are dicts.

        Returns:
            ufl.algebra.Product: the diffusion coefficient
        """
        # TODO use get_D_0 and get_E_D to refactore this method, something like:
        # D_0 = self.get_D_0(species=species)
        # E_D = self.get_E_D(species=species)

        # D_0 = as_fenics_constant(D_0, mesh)
        # E_D = as_fenics_constant(E_D, mesh)

        # return D_0 * ufl.exp(-E_D / k_B / temperature)

        if self.D:
            assert isinstance(self.D, fem.Function)
            return self.D

        if isinstance(self.D_0, float | int) and isinstance(self.E_D, float | int):
            D_0 = as_fenics_constant(self.D_0, mesh)
            E_D = as_fenics_constant(self.E_D, mesh)

            return D_0 * ufl.exp(-E_D / k_B / temperature)

        elif isinstance(self.D_0, dict) and isinstance(self.E_D, dict):
            # check D_0 and E_D have the same keys
            # this check should go in a setter
            if list(self.D_0.keys()) != list(self.E_D.keys()):
                raise ValueError("D_0 and E_D have different keys")

            if species is None:
                raise ValueError("species must be provided if D_0 and E_D are dicts")

            if species in self.D_0:
                D_0 = as_fenics_constant(self.D_0[species], mesh)
            elif species.name in self.D_0:
                D_0 = as_fenics_constant(self.D_0[species.name], mesh)
            else:
                raise ValueError(f"{species} is not in D_0 keys")

            if species in self.E_D:
                E_D = as_fenics_constant(self.E_D[species], mesh)
            elif species.name in self.E_D:
                E_D = as_fenics_constant(self.E_D[species.name], mesh)

            return D_0 * ufl.exp(-E_D / k_B / temperature)

        else:
            raise ValueError("D_0 and E_D must be either floats or dicts")

    def get_solubility_coefficient(self, mesh, temperature, species=None):
        """Defines the solubility coefficient

        Args:

            mesh (dolfinx.mesh.Mesh): the domain mesh
            temperature (dolfinx.fem.Constant): the temperature
            species (festim.Species, optional): the species we want the solubility
                coefficient of. Only needed if K_S_0 and E_K_S are dicts.

        Returns:
            ufl.algebra.Product: the solubility coefficient
        """
        # TODO use get_K_S0 and get_E_K_S to refactore this method, something like:
        # K_S_0 = self.get_K_S_0(species=species)
        # E_K_S = self.get_E_K_S(species=species)

        # K_S_0 = as_fenics_constant(K_S_0, mesh)
        # E_K_S = as_fenics_constant(E_K_S, mesh)

        # return K_S_0 * ufl.exp(-E_K_S / k_B / temperature)

        if isinstance(self.K_S_0, float | int) and isinstance(self.E_K_S, float | int):
            K_S_0 = as_fenics_constant(self.K_S_0, mesh)
            E_K_S = as_fenics_constant(self.E_K_S, mesh)

            return K_S_0 * ufl.exp(-E_K_S / k_B / temperature)

        elif isinstance(self.K_S_0, dict) and isinstance(self.E_K_S, dict):
            # check D_0 and E_D have the same keys
            # this check should go in a setter
            if list(self.K_S_0.keys()) != list(self.E_K_S.keys()):
                raise ValueError("K_S_0 and E_K_S have different keys")

            if species is None:
                raise ValueError(
                    "species must be provided if K_S_0 and E_K_S are dicts"
                )

            if species in self.K_S_0:
                K_S_0 = as_fenics_constant(self.K_S_0[species], mesh)
            elif species.name in self.K_S_0:
                K_S_0 = as_fenics_constant(self.K_S_0[species.name], mesh)
            else:
                raise ValueError(f"{species} is not in K_S_0 keys")

            if species in self.E_K_S:
                E_K_S = as_fenics_constant(self.E_K_S[species], mesh)
            elif species.name in self.E_K_S:
                E_K_S = as_fenics_constant(self.E_K_S[species.name], mesh)

            return K_S_0 * ufl.exp(-E_K_S / k_B / temperature)

        else:
            raise ValueError("K_S_0 and E_K_S must be either floats or dicts")



import dolfinx
import numpy as np
import pytest

import festim as F

from .test_multi_mat_penalty import generate_mesh


def test_petsc_options():
    my_model = F.HydrogenTransportProblem(
        petsc_options={
            "ksp_type": "preonly",
            "pc_type": "lu",
            "pc_factor_mat_solver_type": "mumps",
            "coucoucou": 3,
        }
    )

    tungsten = F.Material(D_0=1, E_D=0, K_S_0=1, E_K_S=0)

    my_model.mesh = F.Mesh1D(vertices=np.linspace(0, 1, num=100))
    vol1 = F.VolumeSubdomain1D(id=1, material=tungsten, borders=[0, 1])
    surface1 = F.SurfaceSubdomain1D(id=4, x=0)
    surface2 = F.SurfaceSubdomain1D(id=5, x=1)

    my_model.subdomains = [vol1, surface1, surface2]

    mobile = F.Species(name="H", mobile=True)

    my_model.species = [mobile]

    my_model.temperature = 600

    my_model.boundary_conditions = [
        F.FixedConcentrationBC(subdomain=surface1, species=mobile, value=1),
        F.FixedConcentrationBC(subdomain=surface2, species=mobile, value=0),
    ]

    my_model.settings = F.Settings(atol=1e-6, rtol=1e-6, transient=False)

    my_model.initialise()
    my_model.run()


def test_D_global_on_2d_mesh():
    """
    Test that the D_global is defined correctly on a 2D mesh with two different
    materials. The D_global should be defined as a piecewise constant function
    with two different values, one for each material.
    """
    mesh, mt, ct = generate_mesh(20)

    my_model = F.HydrogenTransportProblem()
    my_model.mesh = F.Mesh(mesh)
    my_model.volume_meshtags = ct
    my_model.facet_meshtags = mt

    material_top = F.Material(D_0=5.0, E_D=0)
    material_bottom = F.Material(D_0=1.0, E_D=0)

    top_domain = F.VolumeSubdomain(4, material=material_top)
    bottom_domain = F.VolumeSubdomain(3, material=material_bottom)

    top_surface = F.SurfaceSubdomain(id=1)
    bottom_surface = F.SurfaceSubdomain(id=2)
    my_model.subdomains = [
        bottom_domain,
        top_domain,
        top_surface,
        bottom_surface,
    ]

    H = F.Species("H", mobile=True)
    my_model.species = [H]

    my_model.boundary_conditions = []

    my_model.temperature = 500.0

    my_model.settings = F.Settings(atol=1e-10, rtol=1e-10, transient=False)

    my_model.initialise()

    # run

    D, _ = my_model.define_D_global(species=H)

    # test that there are two values in D
    assert len(np.unique(D.x.array[:])) == 2


@pytest.mark.parametrize(
    "species",
    [
        [F.Species("H", mobile=True)],
        [F.Species("H", mobile=True), F.Species("D", mobile=True)],
    ],
)
def test_min_max_vol_on_2d_mesh(species):
    """Added test that catches bug #908"""
    mesh, mt, ct = generate_mesh(10)

    my_model = F.HydrogenTransportProblem()
    my_model.mesh = F.Mesh(mesh)
    my_model.volume_meshtags = ct
    my_model.facet_meshtags = mt

    material_bottom = F.Material(D_0=1, E_D=0, K_S_0=1, E_K_S=0)
    material_top = F.Material(D_0=1, E_D=0, K_S_0=1, E_K_S=0)

    top_domain = F.VolumeSubdomain(4, material=material_top)
    bottom_domain = F.VolumeSubdomain(3, material=material_bottom)

    top_surface = F.SurfaceSubdomain(id=1)
    bottom_surface = F.SurfaceSubdomain(id=2)
    my_model.subdomains = [
        bottom_domain,
        top_domain,
        top_surface,
        bottom_surface,
    ]

    H = species[0]

    my_model.species = species

    my_model.boundary_conditions = [
        F.FixedConcentrationBC(top_surface, value=1, species=H),
        F.FixedConcentrationBC(bottom_surface, value=0, species=H),
    ]

    my_model.temperature = 500.0

    my_model.settings = F.Settings(atol=1e-10, rtol=1e-10, transient=False)
    max_top = F.MaximumVolume(field=H, volume=top_domain)
    max_bottom = F.MaximumVolume(field=H, volume=bottom_domain)
    max_top_boundary = F.MaximumSurface(field=H, surface=top_surface)
    max_bottom_boundary = F.MaximumSurface(field=H, surface=bottom_surface)
    min_top = F.MinimumVolume(field=H, volume=top_domain)
    min_bottom = F.MinimumVolume(field=H, volume=bottom_domain)
    min_top_boundary = F.MinimumSurface(field=H, surface=top_surface)
    min_bottom_boundary = F.MinimumSurface(field=H, surface=bottom_surface)
    my_model.exports = [
        max_top,
        max_bottom,
        max_top_boundary,
        max_bottom_boundary,
        min_top,
        min_bottom,
        min_top_boundary,
        min_bottom_boundary,
    ]
    if len(species) == 2:
        my_model.boundary_conditions.append(
            F.FixedConcentrationBC(top_surface, value=0, species=species[1])
        )
        my_model.boundary_conditions.append(
            F.FixedConcentrationBC(bottom_surface, value=1, species=species[1])
        )
        my_model.exports += [
            F.MaximumVolume(field=species[1], volume=top_domain),
            F.MaximumVolume(field=species[1], volume=bottom_domain),
            F.MaximumSurface(field=species[1], surface=top_surface),
            F.MaximumSurface(field=species[1], surface=bottom_surface),
            F.MinimumVolume(field=species[1], volume=top_domain),
            F.MinimumVolume(field=species[1], volume=bottom_domain),
            F.MinimumSurface(field=species[1], surface=top_surface),
            F.MinimumSurface(field=species[1], surface=bottom_surface),
        ]

    my_model.initialise()
    my_model.run()

    assert max_top.value != max_bottom.value
    assert max_top_boundary.value != max_bottom_boundary.value
    assert min_top.value != min_bottom.value
    assert min_top_boundary.value != min_bottom_boundary.value

    if len(species) == 2:
        assert my_model.exports[-1].value != my_model.exports[-2].value
        assert my_model.exports[-1].value != min_bottom_boundary.value


def test_temp_dependent_bc_mixed_domain_temperature_as_function():
    """Test to catch bug 986"""
    mesh, mt, ct = generate_mesh(8)

    V = dolfinx.fem.functionspace(mesh, ("CG", 1))
    T = dolfinx.fem.Function(V)
    T.interpolate(lambda x: 800.0 + 100 * x[0] + x[1])

    my_model = F.HydrogenTransportProblemDiscontinuous()
    my_model.mesh = F.Mesh(mesh)
    my_model.volume_meshtags = ct
    my_model.facet_meshtags = mt

    material_top = F.Material(D_0=1, E_D=0, K_S_0=1, E_K_S=0)
    material_bottom = F.Material(D_0=1, E_D=0, K_S_0=1, E_K_S=0)

    top_domain = F.VolumeSubdomain(4, material=material_top)
    bottom_domain = F.VolumeSubdomain(3, material=material_bottom)

    top_surface = F.SurfaceSubdomain(id=1)
    bottom_surface = F.SurfaceSubdomain(id=2)
    my_model.subdomains = [
        bottom_domain,
        top_domain,
        top_surface,
        bottom_surface,
    ]

    my_model.surface_to_volume = {
        top_surface: top_domain,
        bottom_surface: bottom_domain,
    }

    H = F.Species("H", mobile=True, subdomains=[bottom_domain, top_domain])

    my_model.species = [H]

    my_model.boundary_conditions = [
        F.FixedConcentrationBC(top_surface, value=lambda T: 2 * T, species=H),
        F.FixedConcentrationBC(bottom_surface, value=0, species=H),
    ]

    my_model.temperature = T

    my_model.settings = F.Settings(atol=1e-10, rtol=1e-10, transient=False)

    my_model.initialise()
    my_model.run()

    expected_value = 2 * T.x.array[:].max()
    computed_value = (
        H.subdomain_to_post_processing_solution[top_domain].x.array[:].max()
    )
    assert np.isclose(computed_value, expected_value)


def test_del():
    """Test that deleting a model with exports works fine (no segfault)"""
    my_model = F.HydrogenTransportProblemDiscontinuous()
    H = F.Species("H")
    my_model.exports = [F.VTXSpeciesExport(filename="h.bp", field=H)]

    my_model.__del__()


def test_multispecies_with_immobile():
    """test to catch bug 1036, that the diffusion coefficent for a non-mobile species
    is not requested in the dict given to material"""

    my_model = F.HydrogenTransportProblem()
    my_model.mesh = F.Mesh1D(vertices=np.linspace(0, 1, num=50))

    H = F.Species("H")
    D = F.Species("D")
    trap = F.Species("trap", mobile=False)
    trapping_site = F.ImplicitSpecies(n=100, others=[H])
    my_model.species = [H, trap]

    my_mat = F.Material(
        D_0={H: 1, D: 1},
        E_D={H: 0, D: 0},
    )
    vol = F.VolumeSubdomain1D(id=1, borders=[0, 1], material=my_mat)
    my_model.subdomains = [vol]

    my_model.reactions = [
        F.Reaction(
            reactant=[H, trapping_site],
            k_0=1,
            E_k=0,
            p_0=1,
            E_p=0,
            product=[trap],
            volume=vol,
        )
    ]

    my_model.temperature = 500

    my_model.settings = F.Settings(transient=False, atol=1e-10, rtol=1e-10)

    my_model.initialise()



import ufl
from dolfinx import fem

from festim import k_B
from festim.boundary_conditions import ParticleFluxBC


class SurfaceReactionBCpartial(ParticleFluxBC):
    """Boundary condition representing a surface reaction
    A + B <-> C
    where A, B are the reactants and C is the product
    the forward reaction rate is K_r = k_r0 * exp(-E_kr / (k_B * T))
    and the backward reaction rate is K_d = k_d0 * exp(-E_kd / (k_B * T))
    The reaction rate is:
    K = K_r * C_A * C_B - K_d * P_C
    with C_A, C_B the concentration of species A and B,
    P_C the partial pressure of species C at the surface.

    This class is used to create the flux of a single species entering the surface
    Example: The flux of species A entering the surface is K.


    Args:
        reactant (list): list of F.Species objects representing the reactants
        gas_pressure (float, callable): the partial pressure of the product species
        k_r0 (float): the pre-exponential factor of the forward reaction rate
        E_kr (float): the activation energy of the forward reaction rate (eV)
        k_d0 (float): the pre-exponential factor of the backward reaction rate
        E_kd (float): the activation energy of the backward reaction rate (eV)
        subdomain (F.SurfaceSubdomain): the surface subdomain where the reaction occurs
        species (F.Species): the species to which the flux is applied
    """

    def __init__(
        self,
        reactant,
        gas_pressure,
        k_r0,
        E_kr,
        k_d0,
        E_kd,
        subdomain,
        species,
    ):
        self.reactant = reactant
        self.gas_pressure = gas_pressure
        self.k_r0 = k_r0
        self.E_kr = E_kr
        self.k_d0 = k_d0
        self.E_kd = E_kd
        super().__init__(subdomain=subdomain, value=None, species=species)

    def create_value_fenics(self, mesh, temperature, t: fem.Constant):
        kr = self.k_r0 * ufl.exp(-self.E_kr / (k_B * temperature))
        kd = self.k_d0 * ufl.exp(-self.E_kd / (k_B * temperature))
        if callable(self.gas_pressure):
            gas_pressure = self.gas_pressure(t=t)
        else:
            gas_pressure = self.gas_pressure

        product_of_reactants = self.reactant[0].concentration
        for reactant in self.reactant[1:]:
            product_of_reactants *= reactant.concentration

        self.value_fenics = kd * gas_pressure - kr * product_of_reactants


class SurfaceReactionBC:
    """Boundary condition representing a surface reaction
    A + B <-> C
    where A, B are the reactants and C is the product
    the forward reaction rate is K_r = k_r0 * exp(-E_kr / (k_B * T))
    and the backward reaction rate is K_d = k_d0 * exp(-E_kd / (k_B * T))
    The reaction rate is:
    K = K_r * C_A * C_B - K_d * P_C
    with C_A, C_B the concentration of species A and B,
    P_C the partial pressure of species C at the surface.

    The flux of species A entering the surface is K.
    In the special case where A=B, then the flux of particle entering the surface is 2*K


    Args:
        reactant (list): list of F.Species objects representing the reactants
        gas_pressure (float, callable): the partial pressure of the product species
        k_r0 (float): the pre-exponential factor of the forward reaction rate
        E_kr (float): the activation energy of the forward reaction rate (eV)
        k_d0 (float): the pre-exponential factor of the backward reaction rate
        E_kd (float): the activation energy of the backward reaction rate (eV)
        subdomain (F.SurfaceSubdomain): the surface subdomain where the reaction occurs
    """

    def __init__(
        self,
        reactant,
        gas_pressure,
        k_r0,
        E_kr,
        k_d0,
        E_kd,
        subdomain,
    ):
        self.reactant = reactant
        self.gas_pressure = gas_pressure
        self.k_r0 = k_r0
        self.E_kr = E_kr
        self.k_d0 = k_d0
        self.E_kd = E_kd
        self.subdomain = subdomain

        # create the flux boundary condition for each reactant
        self.flux_bcs = [
            SurfaceReactionBCpartial(
                reactant=self.reactant,
                gas_pressure=self.gas_pressure,
                k_r0=self.k_r0,
                E_kr=self.E_kr,
                k_d0=self.k_d0,
                E_kd=self.E_kd,
                subdomain=self.subdomain,
                species=species,
            )
            for species in self.reactant
        ]

    @property
    def time_dependent(self):
        return False  # no need to update if only using ufl.conditional objects



from mpi4py import MPI

import dolfinx
import numpy as np

from festim.exports.surface_quantity import SurfaceQuantity


class MinimumSurface(SurfaceQuantity):
    """Computes the minimum value of a field on a given surface

    Args:
        field (festim.Species): species for which the minimum surface is computed
        surface (festim.SurfaceSubdomain): surface subdomain
        filename (str, optional): name of the file to which the minimum surface is exported

    Attributes:
        see `festim.SurfaceQuantity`
    """

    @property
    def title(self):
        return f"Minimum {self.field.name} surface {self.surface.id}"

    def compute(self):
        """
        Computes the minimum value of the field on the defined surface
        subdomain, and appends it to the data list
        """
        solution = self.field.post_processing_solution
        entities = self.facet_meshtags.find(self.surface.id)
        if isinstance(solution, dolfinx.fem.Function):
            V = solution.function_space
        else:
            V = self.field.sub_function_space
        mesh = V.mesh
        dofs = dolfinx.fem.locate_dofs_topological(
            V=V, entity_dim=mesh.topology.dim - 1, entities=entities
        )

        self.value = mesh.comm.allreduce(np.min(solution.x.array[dofs]), op=MPI.MIN)
        self.data.append(self.value)



from mpi4py import MPI

import basix
import dolfinx
import numpy as np
import ufl
from dolfinx import fem
from dolfinx.mesh import create_unit_square

import festim as F

from .tools import error_L2


def test_MMS_1_species_1_trap_with_advection():
    """MMS coupled heat and hydrogen test with 1 mobile species and 1 trap in a 1s
    transient, the values of the temperature, mobile and trapped solutions at the last
    time step is compared to an analytical solution"""

    test_mesh_2d = create_unit_square(MPI.COMM_WORLD, 200, 200)
    x_2d = ufl.SpatialCoordinate(test_mesh_2d)

    # coupled simulation properties
    D_0, E_D = 1.2, 0.1
    k_0, E_k = 2.2, 0.6
    p_0, E_p = 0.5, 0.1
    n_trap = 100
    k_B = F.k_B

    # common festim objects
    test_mat = F.Material(D_0=D_0, E_D=E_D)
    test_vol_sub = F.VolumeSubdomain(id=1, material=test_mat)

    boundary = F.SurfaceSubdomain(id=1)
    test_mobile = F.Species("mobile", mobile=True)
    test_trapped = F.Species(name="trapped", mobile=True)
    empty_trap = F.ImplicitSpecies(n=n_trap, others=[test_trapped])

    V = fem.functionspace(test_mesh_2d, ("Lagrange", 1))
    T = fem.Function(V)
    T_expr = lambda x: 100 + 200 * x[0] + 100 * x[1]
    T.interpolate(T_expr)

    # create velocity field
    v_cg = basix.ufl.element(
        "Lagrange",
        test_mesh_2d.topology.cell_name(),
        2,
        shape=(test_mesh_2d.geometry.dim,),
    )
    V_velocity = fem.functionspace(test_mesh_2d, v_cg)
    u = fem.Function(V_velocity)

    def velocity_func(x):
        values = np.zeros((2, x.shape[1]))  # Initialize with zeros

        scalar_value = x[1] * (x[1] - 1)  # Compute the scalar function
        values[0] = scalar_value  # Assign to first component
        values[1] = 0  # Second component remains zero

        return values

    u.interpolate(velocity_func)

    # define hydrogen problem
    exact_mobile_solution = lambda x: 200 * x[0] ** 2 + 300 * x[1] ** 2
    exact_trapped_solution = lambda x: 10 * x[0] ** 2 + 10 * x[1] ** 2

    D = D_0 * ufl.exp(-E_D / (k_B * T))
    k = k_0 * ufl.exp(-E_k / (k_B * T))
    p = p_0 * ufl.exp(-E_p / (k_B * T))

    f = (
        -ufl.div(D * ufl.grad(exact_mobile_solution(x_2d)))
        + ufl.inner(u, ufl.grad(exact_mobile_solution(x_2d)))
        + k * exact_mobile_solution(x_2d) * (n_trap - exact_trapped_solution(x_2d))
        - p * exact_trapped_solution(x_2d)
    )

    g = (
        -ufl.div(D * ufl.grad(exact_trapped_solution(x_2d)))
        - k * exact_mobile_solution(x_2d) * (n_trap - exact_trapped_solution(x_2d))
        + p * exact_trapped_solution(x_2d)
    )

    my_bcs = []
    for species, value in zip(
        [test_mobile, test_trapped], [exact_mobile_solution, exact_trapped_solution]
    ):
        my_bcs.append(
            F.FixedConcentrationBC(subdomain=boundary, value=value, species=species)
        )

    test_hydrogen_problem = F.HydrogenTransportProblem(
        mesh=F.Mesh(test_mesh_2d),
        subdomains=[test_vol_sub, boundary],
        boundary_conditions=my_bcs,
        species=[test_mobile, test_trapped],
        temperature=T,
        reactions=[
            F.Reaction(
                reactant=[test_mobile, empty_trap],
                product=test_trapped,
                k_0=k_0,
                E_k=E_k,
                p_0=p_0,
                E_p=E_p,
                volume=test_vol_sub,
            )
        ],
        sources=[
            F.ParticleSource(value=f, volume=test_vol_sub, species=test_mobile),
            F.ParticleSource(value=g, volume=test_vol_sub, species=test_trapped),
        ],
        advection_terms=[
            F.AdvectionTerm(velocity=u, subdomain=test_vol_sub, species=test_mobile)
        ],
        settings=F.Settings(
            atol=1e-10,
            rtol=1e-10,
            transient=False,
        ),
    )

    test_hydrogen_problem.initialise()
    test_hydrogen_problem.run()

    # compare computed values with exact solutions
    mobile_computed = test_mobile.post_processing_solution
    trapped_computed = test_trapped.post_processing_solution

    L2_error_mobile = error_L2(mobile_computed, exact_mobile_solution)
    L2_error_trapped = error_L2(trapped_computed, exact_trapped_solution)

    assert L2_error_mobile < 2e-03
    assert L2_error_trapped < 7e-05


def test_multi_material_with_advection():
    def generate_mesh(n=20):
        def bottom_boundary(x):
            return np.isclose(x[1], 0.0)

        def top_boundary(x):
            return np.isclose(x[1], 1.0)

        def bottom_left_boundary(x):
            return np.logical_and(np.isclose(x[0], 0.0), x[1] <= 0.5 + 1e-14)

        def half(x):
            return x[1] <= 0.5 + 1e-14

        mesh = dolfinx.mesh.create_unit_square(
            MPI.COMM_WORLD, n, n, dolfinx.mesh.CellType.triangle
        )

        # Split domain in half and set an interface tag of 5
        tdim = mesh.topology.dim
        fdim = tdim - 1
        top_facets = dolfinx.mesh.locate_entities_boundary(mesh, fdim, top_boundary)
        bottom_facets = dolfinx.mesh.locate_entities_boundary(
            mesh, fdim, bottom_boundary
        )
        bot_left_facets = dolfinx.mesh.locate_entities_boundary(
            mesh, fdim, bottom_left_boundary
        )
        num_facets_local = (
            mesh.topology.index_map(fdim).size_local
            + mesh.topology.index_map(fdim).num_ghosts
        )
        facets = np.arange(num_facets_local, dtype=np.int32)
        values = np.full_like(facets, 0, dtype=np.int32)
        values[top_facets] = 1
        values[bottom_facets] = 2
        values[bot_left_facets] = 3

        bottom_cells = dolfinx.mesh.locate_entities(mesh, tdim, half)
        num_cells_local = (
            mesh.topology.index_map(tdim).size_local
            + mesh.topology.index_map(tdim).num_ghosts
        )
        cells = np.full(num_cells_local, 5, dtype=np.int32)
        cells[bottom_cells] = 4
        ct = dolfinx.mesh.meshtags(
            mesh, tdim, np.arange(num_cells_local, dtype=np.int32), cells
        )
        all_b_facets = dolfinx.mesh.compute_incident_entities(
            mesh.topology, ct.find(4), tdim, fdim
        )
        all_t_facets = dolfinx.mesh.compute_incident_entities(
            mesh.topology, ct.find(5), tdim, fdim
        )
        interface = np.intersect1d(all_b_facets, all_t_facets)
        values[interface] = 6

        mt = dolfinx.mesh.meshtags(mesh, mesh.topology.dim - 1, facets, values)
        return mesh, mt, ct

    mesh, mt, ct = generate_mesh()

    # create velocity field
    def create_velocity_field():
        def velocity_func(x):
            values = np.zeros((2, x.shape[1]))  # Initialize with zeros
            scalar_value = -500 * x[1] * (x[1] - 0.5)  # Compute the scalar function
            values[0] = scalar_value  # Assign to first component
            values[1] = 0  # Second component remains zero
            return values

        mesh2, mt2, ct2 = generate_mesh(n=10)
        submesh, cell_map, v_map = dolfinx.mesh.create_submesh(
            mesh2, ct2.dim, ct2.find(4)
        )[0:3]
        v_cg = basix.ufl.element(
            "Lagrange", submesh.topology.cell_name(), 2, shape=(submesh.geometry.dim,)
        )
        V_velocity = dolfinx.fem.functionspace(submesh, v_cg)
        u = dolfinx.fem.Function(V_velocity)
        u.interpolate(velocity_func)
        return u

    u = create_velocity_field()

    my_model = F.HydrogenTransportProblemDiscontinuous()
    my_model.mesh = F.Mesh(mesh)
    my_model.volume_meshtags = ct
    my_model.facet_meshtags = mt

    material_bottom = F.Material(D_0=1, E_D=0 * 0.1)
    material_top = F.Material(D_0=1, E_D=0 * 0.1)

    material_bottom.K_S_0 = 2.0
    material_bottom.E_K_S = 0
    material_top.K_S_0 = 600
    material_top.E_K_S = 0

    top_domain = F.VolumeSubdomain(5, material=material_top)
    bottom_domain = F.VolumeSubdomain(4, material=material_bottom)

    top_surface = F.SurfaceSubdomain(id=1)
    bottom_surface = F.SurfaceSubdomain(id=2)
    inlet_surf = F.SurfaceSubdomain(id=3)
    my_model.subdomains = [
        bottom_domain,
        top_domain,
        top_surface,
        bottom_surface,
        inlet_surf,
    ]

    # we should be able to automate this
    my_model.interfaces = [F.Interface(6, (bottom_domain, top_domain))]
    my_model.surface_to_volume = {
        top_surface: top_domain,
        bottom_surface: bottom_domain,
        inlet_surf: bottom_domain,
    }

    H = F.Species("H", mobile=True)

    my_model.species = [H]

    for species in my_model.species:
        species.subdomains = [bottom_domain, top_domain]

    my_model.boundary_conditions = [
        F.FixedConcentrationBC(top_surface, value=0.0, species=H),
        F.FixedConcentrationBC(inlet_surf, value=1.0, species=H),
        F.FixedConcentrationBC(bottom_surface, value=0.0, species=H),
    ]

    my_model.advection_terms = [
        F.AdvectionTerm(velocity=u, species=[H], subdomain=bottom_domain)
    ]

    my_model.temperature = 500.0

    my_model.settings = F.Settings(atol=1e-10, rtol=1e-10, transient=False)

    my_model.initialise()
    my_model.run()



# Configuration file for the Sphinx documentation builder.
#
# For the full list of built-in configuration values, see the documentation:
# https://www.sphinx-doc.org/en/master/usage/configuration.html

# -- Path setup --------------------------------------------------------------

# If extensions (or modules to document with autodoc) are in another directory,
# add these directories to sys.path here. If the directory is relative to the
# documentation root, use os.path.abspath to make it absolute, like shown here.
#
import os
import sys

sys.path.insert(0, os.path.abspath("../../src"))


# Generate the map

# Add the directory containing your Python script to the Python path
sys.path.insert(0, os.path.abspath("."))

import map

m = map.generate_map()
current_dir = os.path.dirname(__file__)
html_path = os.path.join(current_dir, "_static", "map.html")

# create _static directory if it doesn't exist
os.makedirs(os.path.dirname(html_path), exist_ok=True)

m.save(html_path)

# -- Project information -----------------------------------------------------
# https://www.sphinx-doc.org/en/master/usage/configuration.html#project-information

project = "FESTIM"
copyright = "2022-2023, FESTIM contributors"
author = "FESTIM-dev"
release = "1.0.0"
version = "1.0"


# -- General configuration ---------------------------------------------------

# Add any Sphinx extension module names here, as strings. They can be
# extensions coming with Sphinx (named 'sphinx.ext.*') or your custom
# ones.
extensions = [
    "sphinx.ext.autodoc",
    "sphinx.ext.autosummary",
    "sphinx.ext.autosectionlabel",
    "sphinx.ext.doctest",
    "sphinx.ext.intersphinx",
    "sphinx.ext.coverage",
    "sphinx.ext.napoleon",
    "sphinx.ext.viewcode",
    "sphinx_copybutton",
    "sphinx_design",
    "sphinxcontrib.bibtex",
    "matplotlib.sphinxext.plot_directive",
    "sphinxcontrib.images",
]

suppress_warnings = ["autosectionlabel.*"]

napoleon_use_ivar = True  # needed to correctly format class attributes

# Add any paths that contain templates here, relative to this directory.
templates_path = ["_templates"]

# List of patterns, relative to source directory, that match files and
# directories to ignore when looking for source files.
# This pattern also affects html_static_path and html_extra_path.
exclude_patterns = []


source_suffix = ".rst"

# The master toctree document.
master_doc = "index"

# The name of the Pygments (syntax highlighting) style to use.
pygments_style = None


# shorten module names in readme
add_module_names = False

# bibliography file
bibtex_bibfiles = ["bibliography/references.bib"]

# -- Options for HTML output -------------------------------------------------

# The theme to use for HTML and HTML Help pages.  See the documentation for
# a list of builtin themes.
#
html_theme = "sphinx_book_theme"
html_static_path = ["_static"]
html_css_files = ["style.css"]

html_theme_options = {
    "repository_url": "https://github.com/festim-dev/FESTIM",
    "use_repository_button": True,
    "use_edit_page_button": True,
    "repository_branch": "fenicsx",
    "path_to_docs": "./docs/source",
    "icon_links": [
        {
            "name": "PyPI",
            "url": "https://pypi.org/project/FESTIM/",
            "icon": "https://img.shields.io/pypi/dw/festim",
            "type": "url",
        },
        {
            "name": "Support Forum",
            "url": "https://festim.discourse.group/",
            "icon": "fa-brands fa-discourse",
        },
        {
            "name": "Slack",
            "url": "https://join.slack.com/t/festim-dev/shared_invite/zt-246hw8d6o-htWASLsbdosUo_2nRKCf9g",
            "icon": "fa-brands fa-slack",
        },
    ],
    "article_header_end": [
        "navbar-icon-links",
        "article-header-buttons",
    ],
}

html_sidebars = {
    "**": [
        "navbar-logo",
        "search-button-field",
        "sbt-sidebar-nav",
    ],
}

html_title = "FESTIM Documentation"



import ufl
from scifem import assemble_scalar

from festim.exports.volume_quantity import VolumeQuantity


class TotalVolume(VolumeQuantity):
    """Computes the total value of a field in a given volume

    Args:
        field (festim.Species): species for which the total volume is computed
        volume (festim.VolumeSubdomain): volume subdomain
        filename (str, optional): name of the file to which the total volume is exported

    Attributes:
        see `festim.VolumeQuantity`
    """

    @property
    def title(self):
        return f"Total {self.field.name} volume {self.volume.id}"

    def compute(self, u, dx: ufl.Measure, entity_maps=None):
        """
        Computes the value of the total volume of the field in the volume subdomain
        and appends it to the data list

        Args:
            u: field for which the total volume is computed
            dx: volume measure of the model
            entity_maps: entity maps relating parent mesh and submesh
        """
        self.value = assemble_scalar(u * dx(self.volume.id), entity_maps=entity_maps)
        self.data.append(self.value)



import numpy as np

import festim as F

from .tools import error_L2


def test_run_MMS_cylindrical():
    """
    Tests that festim produces the correct concentration field in cylindrical
    coordinates
    """

    my_mesh = F.Mesh1D(vertices=np.linspace(1, 2, 500), coordinate_system="cylindrical")

    u_exact = lambda x: 1 + x[0] ** 2

    f = -4

    my_mat = F.Material(D_0=1.0, E_D=0)

    left = F.SurfaceSubdomain1D(id=1, x=1)
    right = F.SurfaceSubdomain1D(id=2, x=2)
    my_vol = F.VolumeSubdomain1D(id=3, borders=[1, 2], material=my_mat)

    my_subdomains = [my_vol, left, right]

    H = F.Species("H")
    D = F.Species("D")

    my_bcs = [
        F.FixedConcentrationBC(subdomain=left, value=u_exact, species=H),
        F.FixedConcentrationBC(subdomain=right, value=u_exact, species=H),
    ]

    my_temp = 500

    my_sources = [
        F.ParticleSource(value=f, volume=my_vol, species=H),
    ]

    my_settings = F.Settings(
        atol=1e-10,
        rtol=1e-9,
        max_iterations=50,
        transient=False,
    )

    my_sim = F.HydrogenTransportProblem(
        mesh=my_mesh,
        species=[H, D],
        subdomains=my_subdomains,
        boundary_conditions=my_bcs,
        temperature=my_temp,
        sources=my_sources,
        settings=my_settings,
    )

    my_sim.initialise()
    my_sim.run()

    computed_solution = H.post_processing_solution

    L2_error = error_L2(computed_solution, u_exact)

    assert L2_error < 1e-6


def test_run_MMS_cylindrical_mixed_domain():
    """
    Tests that festim produces the correct concentration field in cylindrical
    coordinates in a discontinuous domain with two materials
    """

    my_model = F.HydrogenTransportProblemDiscontinuous()

    r_interface = 2
    left_domain = np.linspace(1, r_interface, num=1000)
    right_domain = np.linspace(r_interface, r_interface + 1, num=1000)

    vertices = np.concatenate(
        [
            left_domain,
            right_domain,
        ]
    )
    my_mesh = F.Mesh1D(vertices=vertices, coordinate_system="cylindrical")

    my_model.mesh = my_mesh

    K_S_left = 3.0
    K_S_right = 2.0
    D = 2.0

    def c_exact_left(x):
        return (r_interface - x[0]) ** 2 + 2

    def c_exact_right(x):
        return K_S_right / K_S_left * c_exact_left(x)

    lap_c = lambda r: 4 - 2 * r_interface / r

    mat_1 = F.Material(D_0=D, E_D=0, K_S_0=K_S_left, E_K_S=0, solubility_law="sievert")
    mat_2 = F.Material(D_0=D, E_D=0, K_S_0=K_S_right, E_K_S=0, solubility_law="sievert")

    left = F.SurfaceSubdomain1D(id=1, x=left_domain[0])
    right = F.SurfaceSubdomain1D(id=2, x=right_domain[-1])
    vol_1 = F.VolumeSubdomain1D(
        id=3, borders=[left_domain[0], left_domain[-1]], material=mat_1
    )
    vol_2 = F.VolumeSubdomain1D(
        id=4, borders=[right_domain[0], right_domain[-1]], material=mat_2
    )

    my_model.subdomains = [vol_1, vol_2, left, right]

    my_model.interfaces = [F.Interface(5, (vol_1, vol_2), penalty_term=100)]
    my_model.surface_to_volume = {
        left: vol_1,
        right: vol_2,
    }

    H = F.Species("H", mobile=True, subdomains=[vol_1, vol_2])
    my_model.species = [H]

    my_model.boundary_conditions = [
        F.FixedConcentrationBC(subdomain=left, value=c_exact_left, species=H),
        F.FixedConcentrationBC(subdomain=right, value=c_exact_right, species=H),
    ]

    my_model.temperature = 500

    f_left = lambda x: -D * lap_c(x[0])
    f_right = lambda x: -D * K_S_right / K_S_left * lap_c(x[0])

    my_model.sources = [
        F.ParticleSource(value=f_left, volume=vol_1, species=H),
        F.ParticleSource(value=f_right, volume=vol_2, species=H),
    ]

    my_model.settings = F.Settings(
        atol=1e-10, rtol=1e-10, max_iterations=10, transient=False, element_degree=1
    )

    my_model.initialise()
    my_model.run()

    c_l_computed = H.subdomain_to_post_processing_solution[vol_1]
    c_r_computed = H.subdomain_to_post_processing_solution[vol_2]

    L2_error_l = error_L2(c_l_computed, c_exact_left)
    L2_error_r = error_L2(c_r_computed, c_exact_right)

    assert L2_error_l < 1e-06
    assert L2_error_r < 1e-06



from scifem import assemble_scalar

from festim.exports.volume_quantity import VolumeQuantity


class AverageVolume(VolumeQuantity):
    """Computes the average value of a field in a given volume

    Args:
        field (festim.Species): species for which the average volume is computed
        volume (festim.VolumeSubdomain): volume subdomain
        filename (str, optional): name of the file to which the average volume is exported

    Attributes:
        see `festim.VolumeQuantity`
    """

    @property
    def title(self):
        return f"Average {self.field.name} volume {self.volume.id}"

    def compute(self, u, dx, entity_maps=None):
        """
        Computes the average value of solution function within the defined volume
        subdomain, and appends it to the data list
        """
        self.value = assemble_scalar(
            u * dx(self.volume.id), entity_maps=entity_maps
        ) / assemble_scalar(1 * dx(self.volume.id))
        self.data.append(self.value)



from mpi4py import MPI

import dolfinx
import dolfinx.fem.petsc
import numpy as np
import ufl
from dolfinx import fem

import festim as F


def generate_mesh(n=20):
    def bottom_boundary(x):
        return np.isclose(x[1], 0.0)

    def top_boundary(x):
        return np.isclose(x[1], 1.0)

    def half(x):
        return x[1] <= 0.5 + 1e-14

    mesh = dolfinx.mesh.create_unit_square(
        MPI.COMM_WORLD, n, n, dolfinx.mesh.CellType.triangle
    )

    # Split domain in half and set an interface tag of 5
    gdim = mesh.geometry.dim
    tdim = mesh.topology.dim
    fdim = tdim - 1
    top_facets = dolfinx.mesh.locate_entities_boundary(mesh, fdim, top_boundary)
    bottom_facets = dolfinx.mesh.locate_entities_boundary(mesh, fdim, bottom_boundary)
    num_facets_local = (
        mesh.topology.index_map(fdim).size_local
        + mesh.topology.index_map(fdim).num_ghosts
    )
    facets = np.arange(num_facets_local, dtype=np.int32)
    values = np.full_like(facets, 0, dtype=np.int32)
    values[top_facets] = 1
    values[bottom_facets] = 2

    bottom_cells = dolfinx.mesh.locate_entities(mesh, tdim, half)
    num_cells_local = (
        mesh.topology.index_map(tdim).size_local
        + mesh.topology.index_map(tdim).num_ghosts
    )
    cells = np.full(num_cells_local, 4, dtype=np.int32)
    cells[bottom_cells] = 3
    ct = dolfinx.mesh.meshtags(
        mesh, tdim, np.arange(num_cells_local, dtype=np.int32), cells
    )
    all_b_facets = dolfinx.mesh.compute_incident_entities(
        mesh.topology, ct.find(3), tdim, fdim
    )
    all_t_facets = dolfinx.mesh.compute_incident_entities(
        mesh.topology, ct.find(4), tdim, fdim
    )
    interface = np.intersect1d(all_b_facets, all_t_facets)
    values[interface] = 5

    mt = dolfinx.mesh.meshtags(mesh, mesh.topology.dim - 1, facets, values)
    return mesh, mt, ct


def test_run(tmpdir):
    my_model = F.HydrogenTransportProblemDiscontinuousChangeVar()

    interface_1 = 0.2
    interface_2 = 0.8

    vertices = np.concatenate(
        [
            np.linspace(0, interface_1, num=100),
            np.linspace(interface_1, interface_2, num=100),
            np.linspace(interface_2, 1, num=100),
        ]
    )

    my_model.mesh = F.Mesh1D(vertices)

    material_left = F.Material(D_0=2.0, E_D=0.1, K_S_0=2.0, E_K_S=0)
    material_mid = F.Material(D_0=2.0, E_D=0.1, K_S_0=4.0, E_K_S=0)
    material_right = F.Material(D_0=2.0, E_D=0.1, K_S_0=6.0, E_K_S=0)

    left_domain = F.VolumeSubdomain1D(
        3, borders=[0, interface_1], material=material_left
    )
    middle_domain = F.VolumeSubdomain1D(
        4, borders=[interface_1, interface_2], material=material_mid
    )
    right_domain = F.VolumeSubdomain1D(
        5, borders=[interface_2, 1], material=material_right
    )

    left_surface = F.SurfaceSubdomain1D(id=1, x=vertices[0])
    right_surface = F.SurfaceSubdomain1D(id=2, x=vertices[-1])

    my_model.subdomains = [
        left_domain,
        middle_domain,
        right_domain,
        left_surface,
        right_surface,
    ]

    H = F.Species("H", mobile=True)
    trapped_H = F.Species("H_trapped", mobile=False)
    empty_trap = F.ImplicitSpecies(n=0.5, others=[trapped_H])

    my_model.species = [H, trapped_H]

    for species in [H, trapped_H]:
        species.subdomains = [left_domain, middle_domain, right_domain]

    my_model.reactions = [
        F.Reaction(
            reactant=[H, empty_trap],
            product=[trapped_H],
            k_0=2,
            E_k=0,
            p_0=0.1,
            E_p=0,
            volume=domain,
        )
        for domain in [left_domain, middle_domain, right_domain]
    ]

    K_left = material_left.K_S_0
    K_right = material_right.K_S_0
    assert material_left.E_K_S == 0, "Only constant solubility is supported"
    assert material_right.E_K_S == 0, "Only constant solubility is supported"

    my_model.boundary_conditions = [
        F.FixedConcentrationBC(left_surface, value=0.05 / K_left, species=H),
        F.FixedConcentrationBC(right_surface, value=0.2 / K_right, species=H),
    ]

    my_model.temperature = lambda x: 300 + 100 * x[0]

    my_model.settings = F.Settings(atol=1e-10, rtol=1e-10, final_time=10)
    my_model.settings.stepsize = F.Stepsize(1)

    my_model.exports = [
        F.VTXSpeciesExport(tmpdir + f"/u_{field}_var_change.bp", field=field)
        for field in [H, trapped_H]
    ]
    my_model.initialise()
    my_model.run()


def test_2_materials_2d_mms(tmpdir):
    """
    MMS case for a 2D problem with 2 materials
    adapted from https://festim-vv-report.readthedocs.io/en/v1.0/verification/mms/discontinuity.html
    """
    K_S_top = 3.0
    K_S_bot = 6.0
    D_top = 2.0
    D_bot = 5.0
    c_exact_top_ufl = (
        lambda x: 1 + ufl.sin(ufl.pi * (2 * x[0] + 0.5)) + ufl.cos(2 * ufl.pi * x[1])
    )

    def c_exact_bot_ufl(x):
        return K_S_bot / K_S_top * c_exact_top_ufl(x)

    c_exact_top_np = (
        lambda x: 1 + np.sin(np.pi * (2 * x[0] + 0.5)) + np.cos(2 * np.pi * x[1])
    )

    def c_exact_bot_np(x):
        return K_S_bot / K_S_top * c_exact_top_np(x)

    mesh, mt, ct = generate_mesh(100)

    my_model = F.HydrogenTransportProblemDiscontinuousChangeVar()
    my_model.mesh = F.Mesh(mesh)
    my_model.volume_meshtags = ct
    my_model.facet_meshtags = mt

    material_bottom = F.Material(D_0=D_bot, E_D=0, K_S_0=K_S_bot, E_K_S=0)
    material_top = F.Material(D_0=D_top, E_D=0, K_S_0=K_S_top, E_K_S=0)

    top_domain = F.VolumeSubdomain(4, material=material_top)
    bottom_domain = F.VolumeSubdomain(3, material=material_bottom)

    top_surface = F.SurfaceSubdomain(id=1)
    bottom_surface = F.SurfaceSubdomain(id=2)
    my_model.subdomains = [
        bottom_domain,
        top_domain,
        top_surface,
        bottom_surface,
    ]

    H = F.Species("H", mobile=True)

    my_model.species = [H]

    my_model.boundary_conditions = [
        F.FixedConcentrationBC(top_surface, value=c_exact_top_ufl, species=H),
        F.FixedConcentrationBC(bottom_surface, value=c_exact_bot_ufl, species=H),
    ]

    source_top_val = (
        lambda x: 8
        * ufl.pi**2
        * (ufl.cos(2 * ufl.pi * x[0]) + ufl.cos(2 * ufl.pi * x[1]))
    )
    source_bottom_val = (
        lambda x: 40
        * ufl.pi**2
        * (ufl.cos(2 * ufl.pi * x[0]) + ufl.cos(2 * ufl.pi * x[1]))
    )
    my_model.sources = [
        F.ParticleSource(volume=top_domain, species=H, value=source_top_val),
        F.ParticleSource(volume=bottom_domain, species=H, value=source_bottom_val),
    ]

    my_model.temperature = 500.0  # lambda x: 300 + 10 * x[1] + 100 * x[0]

    my_model.settings = F.Settings(atol=1e-10, rtol=1e-10, transient=False)
    my_model.exports = [F.VTXSpeciesExport(tmpdir + "/u.bp", field=H)]

    my_model.initialise()
    my_model.run()

    c_computed = H.post_processing_solution

    top_cells = dolfinx.mesh.locate_entities(
        mesh, mesh.topology.dim, lambda x: x[1] > 0.5 - 1e-14
    )
    bot_cells = dolfinx.mesh.locate_entities(
        mesh, mesh.topology.dim, lambda x: x[1] <= 0.5 + 1e-14
    )

    # Create higher order function space
    degree = c_computed.function_space.ufl_element().degree
    family = "DG"
    mesh = c_computed.function_space.mesh
    W = fem.functionspace(mesh, (family, degree + 3))
    # Interpolate approximate solution
    u_W = fem.Function(W)
    u_W.interpolate(c_computed, cells0=top_cells)
    u_W.interpolate(c_computed, cells0=bot_cells)

    # Interpolate exact solution, special handling if exact solution
    # is a ufl expression or a python lambda function

    u_ex_W = fem.Function(W)
    u_ex_W.interpolate(c_exact_top_np, cells0=top_cells)
    u_ex_W.interpolate(c_exact_bot_np, cells0=bot_cells)

    # Compute the error in the higher order function space
    e_W = fem.Function(W)
    e_W.x.array[:] = u_W.x.array - u_ex_W.x.array

    # Integrate the error
    error = fem.form(ufl.inner(e_W, e_W) * ufl.dx)
    error_local = fem.assemble_scalar(error)
    error_global = mesh.comm.allreduce(error_local, op=MPI.SUM)
    L2_error_mobile = np.sqrt(error_global)

    assert L2_error_mobile < 1e-3



import ufl
from dolfinx import fem
from scifem import assemble_scalar

from festim.exports.surface_quantity import SurfaceQuantity
from festim.species import Species
from festim.subdomain.surface_subdomain import SurfaceSubdomain


class SurfaceFlux(SurfaceQuantity):
    """Computes the flux of a field on a given surface

    Args:
        field: species for which the surface flux is computed
        surface: surface subdomain
        filename: name of the file to which the surface flux is exported

    Attributes:
        see `festim.SurfaceQuantity`
    """

    field: Species
    surface: SurfaceSubdomain
    filename: str

    title: str
    value: float
    data: list[float]

    @property
    def title(self):
        return f"{self.field.name} flux surface {self.surface.id}"

    def __init__(
        self, field: Species, surface: SurfaceSubdomain, filename: str | None = None
    ) -> None:
        super().__init__(field=field, surface=surface, filename=filename)

    def compute(
        self, u: fem.Function | ufl.indexed.Indexed, ds: ufl.Measure, entity_maps=None
    ):
        """Computes the value of the flux at the surface

        Args:
            u: field for which the flux is computed
            ds: surface measure of the model
            entity_maps: entity maps relating parent mesh and submesh
        """

        # obtain mesh normal from field
        # if case multispecies, solution is an index, use sub_function_space
        if isinstance(u, ufl.indexed.Indexed):
            mesh = self.field.sub_function_space.mesh
        else:
            mesh = u.function_space.mesh
        n = ufl.FacetNormal(mesh)

        self.value = assemble_scalar(
            fem.form(
                -self.D * ufl.dot(ufl.grad(u), n) * ds(self.surface.id),
                entity_maps=entity_maps,
            )
        )
        self.data.append(self.value)



import basix
import ufl
from dolfinx import fem
from dolfinx.io import VTXWriter

from festim import boundary_conditions, exports, helpers, problem
from festim import source as _source


class HeatTransferProblem(problem.ProblemBase):
    def __init__(
        self,
        mesh=None,
        subdomains=None,
        initial_condition=None,
        boundary_conditions=None,
        sources=None,
        exports=None,
        settings=None,
    ) -> None:
        super().__init__(
            mesh=mesh,
            sources=sources,
            exports=exports,
            subdomains=subdomains,
            boundary_conditions=boundary_conditions,
            settings=settings,
        )

        self.initial_condition = initial_condition

    @property
    def sources(self):
        return self._sources

    @sources.setter
    def sources(self, value):
        if not all(isinstance(source, _source.HeatSource) for source in value):
            raise TypeError("sources must be a list of festim.HeatSource objects")
        self._sources = value

    @property
    def boundary_conditions(self):
        return self._boundary_conditions

    @boundary_conditions.setter
    def boundary_conditions(self, value):
        if not all(
            isinstance(
                bc,
                boundary_conditions.FixedTemperatureBC | boundary_conditions.HeatFluxBC,
            )
            for bc in value
        ):
            raise TypeError(
                "boundary_conditions must be a list of festim.FixedTemperatureBC "
                "or festim.HeatFluxBC objects"
            )
        self._boundary_conditions = value

    def initialise(self):
        self.define_function_space()
        self.define_meshtags_and_measures()

        self.t = fem.Constant(self.mesh.mesh, 0.0)
        if self.settings.transient:
            # TODO should raise error if no stepsize is provided
            # TODO Should this be an attribute of festim.Stepsize?
            self._dt = helpers.as_fenics_constant(
                self.settings.stepsize.initial_value, self.mesh.mesh
            )

        self.define_boundary_conditions()
        self.create_source_values_fenics()
        self.create_flux_values_fenics()
        self.create_initial_conditions()
        self.create_formulation()
        self.create_solver()
        self.initialise_exports()

    def define_function_space(self):
        """Creates the function space of the model, creates a mixed element if
        model is multispecies. Creates the main solution and previous solution
        function u and u_n. Create global DG function spaces of degree 0 and 1
        for the global diffusion coefficient"""

        degree = 1
        element_CG = basix.ufl.element(
            basix.ElementFamily.P,
            self.mesh.mesh.basix_cell(),
            degree,
            basix.LagrangeVariant.equispaced,
        )

        self.function_space = fem.functionspace(self.mesh.mesh, element_CG)

        self.u = fem.Function(self.function_space)
        self.u_n = fem.Function(self.function_space)
        self.test_function = ufl.TestFunction(self.function_space)

    def create_dirichletbc_form(self, bc):
        """Creates a dirichlet boundary condition form

        Args:
            bc (festim.FixedTemperatureBC): the boundary condition

        Returns:
            dolfinx.fem.bcs.DirichletBC: A representation of
                the boundary condition for modifying linear systems.
        """
        bc.create_value(
            function_space=self.function_space,
            t=self.t,
        )

        bc_dofs = bc.define_surface_subdomain_dofs(
            facet_meshtags=self.facet_meshtags,
            function_space=self.function_space,
        )

        if isinstance(bc.value_fenics, (fem.Function)):
            form = fem.dirichletbc(value=bc.value_fenics, dofs=bc_dofs)
        else:
            form = fem.dirichletbc(
                value=bc.value_fenics, dofs=bc_dofs, V=self.function_space
            )

        return form

    def create_source_values_fenics(self):
        """For each source create the value_fenics"""
        for source in self.sources:
            # create value_fenics for all source objects
            source.value.convert_input_value(
                function_space=self.function_space,
                t=self.t,
                up_to_ufl_expr=True,
            )

    def create_flux_values_fenics(self):
        """For each heat flux create the value_fenics"""
        for bc in self.boundary_conditions:
            # create value_fenics for all F.HeatFluxBC objects
            if isinstance(bc, boundary_conditions.HeatFluxBC):
                bc.create_value_fenics(
                    mesh=self.mesh.mesh,
                    temperature=self.u,
                    t=self.t,
                )

    def create_initial_conditions(self):
        """For each initial condition, create the value_fenics and assign it to
        the previous solution of the condition's species"""

        if not self.initial_condition:
            return

        if not self.settings.transient:
            raise ValueError(
                "Initial conditions can only be defined for transient simulations"
            )

        # create value_fenics for condition

        self.initial_condition.create_expr_fenics(
            mesh=self.mesh.mesh,
            function_space=self.function_space,
        )

        # assign to previous solution of species
        self.u_n.interpolate(self.initial_condition.expr_fenics)

    def create_formulation(self):
        """Creates the formulation of the model"""

        self.formulation = 0

        # add diffusion and time derivative for each species
        for vol in self.volume_subdomains:
            thermal_cond = vol.material.thermal_conductivity
            if callable(thermal_cond):
                thermal_cond = thermal_cond(self.u)

            self.formulation += ufl.dot(
                thermal_cond * ufl.grad(self.u), ufl.grad(self.test_function)
            ) * self.dx(vol.id)

            if self.settings.transient:
                density = vol.material.density
                heat_capacity = vol.material.heat_capacity
                if callable(density):
                    density = density(self.u)
                if callable(heat_capacity):
                    heat_capacity = heat_capacity(self.u)
                self.formulation += (
                    density
                    * heat_capacity
                    * ((self.u - self.u_n) / self.dt)
                    * self.test_function
                    * self.dx(vol.id)
                )

        # add sources
        for source in self.sources:
            self.formulation -= (
                source.value.fenics_object
                * self.test_function
                * self.dx(source.volume.id)
            )

        # add fluxes
        for bc in self.boundary_conditions:
            if isinstance(bc, boundary_conditions.HeatFluxBC):
                self.formulation -= (
                    bc.value_fenics * self.test_function * self.ds(bc.subdomain.id)
                )

    def initialise_exports(self):
        """Defines the export writers of the model, if field is given as
        a string, find species object in self.species"""

        for export in self.exports:
            if isinstance(export, exports.XDMFExport):
                raise NotImplementedError(
                    "XDMF export is not implemented yet for heat transfer problems"
                )
            if isinstance(export, exports.VTXTemperatureExport):
                export.writer = VTXWriter(
                    self.u.function_space.mesh.comm,
                    export.filename,
                    [self.u],
                    engine="BP5",
                )

    def post_processing(self):
        """Post processes the model"""

        for export in self.exports:
            # skip if it isn't time to export
            if hasattr(export, "times"):
                if not helpers.is_it_time_to_export(
                    current_time=float(self.t), times=export.times
                ):
                    continue
            # TODO if export type derived quantity
            if isinstance(export, exports.SurfaceQuantity):
                raise NotImplementedError(
                    "SurfaceQuantity export is not implemented yet "
                    "for heat transfer problems"
                )
                export.compute(
                    self.mesh.n,
                    self.ds,
                )
                # update export data
                export.t.append(float(self.t))

                # if filename given write export data to file
                if export.filename is not None:
                    export.write(t=float(self.t))
            if isinstance(export, exports.XDMFExport):
                export.write(float(self.t))

            if isinstance(export, exports.VTXTemperatureExport):
                export.writer.write(float(self.t))

    def __del__(self):
        for export in self.exports:
            if isinstance(export, exports.VTXTemperatureExport):
                export.writer.close()



from mpi4py import MPI

import dolfinx
import dolfinx.fem.petsc
import numpy as np
import ufl

import festim as F


# ---------------- Generate a mesh ----------------
def generate_mesh():
    def bottom_boundary(x):
        return np.isclose(x[1], 0.0)

    def top_boundary(x):
        return np.isclose(x[1], 1.0)

    def half(x):
        return x[1] <= 0.5 + 1e-14

    mesh = dolfinx.mesh.create_unit_square(
        MPI.COMM_WORLD, 20, 20, dolfinx.mesh.CellType.triangle
    )

    # Split domain in half and set an interface tag of 5
    gdim = mesh.geometry.dim
    tdim = mesh.topology.dim
    fdim = tdim - 1
    top_facets = dolfinx.mesh.locate_entities_boundary(mesh, fdim, top_boundary)
    bottom_facets = dolfinx.mesh.locate_entities_boundary(mesh, fdim, bottom_boundary)
    num_facets_local = (
        mesh.topology.index_map(fdim).size_local
        + mesh.topology.index_map(fdim).num_ghosts
    )
    facets = np.arange(num_facets_local, dtype=np.int32)
    values = np.full_like(facets, 0, dtype=np.int32)
    values[top_facets] = 1
    values[bottom_facets] = 2

    bottom_cells = dolfinx.mesh.locate_entities(mesh, tdim, half)
    num_cells_local = (
        mesh.topology.index_map(tdim).size_local
        + mesh.topology.index_map(tdim).num_ghosts
    )
    cells = np.full(num_cells_local, 4, dtype=np.int32)
    cells[bottom_cells] = 3
    ct = dolfinx.mesh.meshtags(
        mesh, tdim, np.arange(num_cells_local, dtype=np.int32), cells
    )
    all_b_facets = dolfinx.mesh.compute_incident_entities(
        mesh.topology, ct.find(3), tdim, fdim
    )
    all_t_facets = dolfinx.mesh.compute_incident_entities(
        mesh.topology, ct.find(4), tdim, fdim
    )
    interface = np.intersect1d(all_b_facets, all_t_facets)
    values[interface] = 5

    mt = dolfinx.mesh.meshtags(mesh, mesh.topology.dim - 1, facets, values)
    return mesh, mt, ct


mesh, mt, ct = generate_mesh()

my_model = F.HydrogenTransportProblemDiscontinuous()
my_model.mesh = F.Mesh(mesh)
my_model.volume_meshtags = ct
my_model.facet_meshtags = mt

material_bottom = F.Material(D_0=2.0, E_D=0.1)
material_top = F.Material(D_0=2.0, E_D=0.1)

material_bottom.K_S_0 = 2.0
material_bottom.E_K_S = 0
material_top.K_S_0 = 4.0
material_top.E_K_S = 0

top_domain = F.VolumeSubdomain(4, material=material_top)
bottom_domain = F.VolumeSubdomain(3, material=material_bottom)

top_surface = F.SurfaceSubdomain(id=1)
bottom_surface = F.SurfaceSubdomain(id=2)
my_model.subdomains = [bottom_domain, top_domain, top_surface, bottom_surface]

# we should be able to automate this
my_model.interfaces = [F.Interface(5, (bottom_domain, top_domain))]
my_model.surface_to_volume = {
    top_surface: top_domain,
    bottom_surface: bottom_domain,
}

H = F.Species("H", mobile=True)
trapped_H = F.Species("H_trapped", mobile=False)
empty_trap = F.ImplicitSpecies(n=0.5, others=[trapped_H])

my_model.species = [H, trapped_H]

for species in my_model.species:
    species.subdomains = [bottom_domain, top_domain]


my_model.reactions = [
    F.Reaction(
        reactant=[H, empty_trap],
        product=[trapped_H],
        k_0=2,
        E_k=0,
        p_0=0.1,
        E_p=0,
        volume=top_domain,
    ),
    F.Reaction(
        reactant=[H, empty_trap],
        product=[trapped_H],
        k_0=2,
        E_k=0,
        p_0=0.1,
        E_p=0,
        volume=bottom_domain,
    ),
]

my_model.boundary_conditions = [
    F.FixedConcentrationBC(top_surface, value=0.05, species=H),
    F.FixedConcentrationBC(bottom_surface, value=0.2, species=H),
]


my_model.temperature = lambda x: 400 + 10 * x[1] + 100 * x[0]

my_model.settings = F.Settings(atol=1e-10, rtol=1e-5, transient=False)

my_model.exports = [
    F.VTXSpeciesExport(f"c_{subdomain.id}.bp", field=H, subdomain=subdomain)
    for subdomain in my_model.volume_subdomains
] + [
    F.VTXSpeciesExport(f"c_t_{subdomain.id}.bp", field=trapped_H, subdomain=subdomain)
    for subdomain in my_model.volume_subdomains
]

my_model.initialise()
my_model.run()

# -------------------- post processing --------------------


# derived quantities


ds = ufl.Measure("ds", domain=mesh, subdomain_data=my_model.facet_meshtags)
ds_b = ufl.Measure("ds", domain=bottom_domain.submesh, subdomain_data=bottom_domain.ft)
ds_t = ufl.Measure("ds", domain=top_domain.submesh, subdomain_data=top_domain.ft)
dx_b = ufl.Measure("dx", domain=bottom_domain.submesh)
dx = ufl.Measure("dx", domain=mesh)

n_b = ufl.FacetNormal(bottom_domain.submesh)
n_t = ufl.FacetNormal(top_domain.submesh)

form = dolfinx.fem.form(bottom_domain.u.sub(0) * dx_b)
print(dolfinx.fem.assemble_scalar(form))

form = dolfinx.fem.form(bottom_domain.u.sub(1) * dx_b)
print(dolfinx.fem.assemble_scalar(form))

# Check for DOLFINx version compatibility
try:
    from dolfinx.mesh import EntityMap  # noqa: F401

    legacy_entity_map = False
    entity_map_wrapper = lambda e_map: list(e_map.values())
except ImportError:
    legacy_entity_map = True
    entity_map_wrapper = lambda e_map: e_map


form = dolfinx.fem.form(
    my_model.temperature_fenics * dx_b,
    entity_maps=entity_map_wrapper({mesh: bottom_domain.cell_map}),
)
print(dolfinx.fem.assemble_scalar(form))

D = bottom_domain.material.get_diffusion_coefficient(
    my_model.mesh.mesh, my_model.temperature_fenics, H
)
id_interface = 5

form = dolfinx.fem.form(
    ufl.dot(
        D * ufl.grad(bottom_domain.u.sub(0)),
        n_b,
    )
    * ds_b(id_interface),
    entity_maps=entity_map_wrapper({mesh: bottom_domain.cell_map}),
)
print(dolfinx.fem.assemble_scalar(form))
form = dolfinx.fem.form(
    ufl.dot(
        D * ufl.grad(top_domain.u.sub(0)),
        n_t,
    )
    * ds_t(id_interface),
    entity_maps=entity_map_wrapper({mesh: top_domain.cell_map}),
)
print(dolfinx.fem.assemble_scalar(form))


# using submesh function
u = H.subdomain_to_post_processing_solution[bottom_domain]
if legacy_entity_map:
    entity_maps = {
        sd.submesh: sd.parent_to_submesh for sd in my_model.volume_subdomains
    }
else:
    entity_maps = [sd.cell_map for sd in my_model.volume_subdomains]

form = dolfinx.fem.form(
    ufl.dot(
        D * ufl.grad(u),
        n_b,
    )
    * ds(bottom_surface.id),
    entity_maps=entity_maps,
)
print(dolfinx.fem.assemble_scalar(form))



from typing import Literal

import festim as F


class Settings:
    """Settings for a festim simulation.

    Args:
        atol (float or callable): Absolute tolerance for the solver.
        rtol (float or callable): Relative tolerance for the solver.
        max_iterations (int, optional): Maximum number of iterations for the
            solver. Defaults to 30.
        transient (bool, optional): Whether the simulation is transient or not.
        final_time (float, optional): Final time for a transient simulation.
            Defaults to None
        element_degree (int, optional): Degree order for finite element.
            Defaults to 1.
        stepsize (festim.Stepsize, optional): stepsize for a transient
            simulation. Defaults to None
        convergence_criterion: resiudal or incremental (for Newton solver)

    Attributes:
        atol (float or callable): Absolute tolerance for the solver.
        rtol (float or callable): Relative tolerance for the solver.
        max_iterations (int): Maximum number of iterations for the solver.
        transient (bool): Whether the simulation is transient or not.
        final_time (float): Final time for a transient simulation.
        element_degree (int): Degree order for finite element.
        stepsize (festim.Stepsize): stepsize for a transient
            simulation.
        convergence_criterion: resiudal or incremental (for Newton solver)

    """

    def __init__(
        self,
        atol,
        rtol,
        max_iterations=30,
        transient=True,
        final_time=None,
        element_degree=1,
        stepsize=None,
        convergence_criterion: Literal["residual", "incremental"] = "residual",
    ) -> None:
        self.atol = atol
        self.rtol = rtol
        self.max_iterations = max_iterations
        self.transient = transient
        self.final_time = final_time
        self.element_degree = element_degree
        self.stepsize = stepsize
        self.convergence_criterion = convergence_criterion

    @property
    def stepsize(self):
        return self._stepsize

    @stepsize.setter
    def stepsize(self, value):
        if value is None:
            self._stepsize = None
        elif isinstance(value, (float, int)):
            self._stepsize = F.Stepsize(initial_value=value)
        elif isinstance(value, F.Stepsize):
            self._stepsize = value
        else:
            raise TypeError("stepsize must be an of type int, float or festim.Stepsize")



import tqdm.autonotebook
from dolfinx import fem

from festim.heat_transfer_problem import HeatTransferProblem
from festim.helpers import nmm_interpolate
from festim.hydrogen_transport_problem import (
    HydrogenTransportProblem,
    HydrogenTransportProblemDiscontinuous,
    HydrogenTransportProblemDiscontinuousChangeVar,
)


class CoupledTransientHeatTransferHydrogenTransport:
    """
    Coupled heat transfer and hydrogen transport transient problem

    Args:
        heat_problem: the heat transfer problem
        hydrogen_problem: the hydrogen transport problem

    Attributes:
        heat_problem: the heat transfer problem
        hydrogen_problem: the hydrogen transport problem
        non_matching_meshes: True if the meshes in the heat_problem and hydrogen_problem
            are not matching

    Examples:
        .. code:: python

            import festim as F

            my_heat_transfer_model = F.HeatTransferProblem(...)

            my_h_transport_model = F.HydrogenTransportProblem(...)

            coupled_problem = F.CoupledTransientHeatTransferHydrogenTransport(
                heat_problem=my_heat_transfer_model,
                hydrogen_problem=my_h_transport_model,
            )


    """

    heat_problem: HeatTransferProblem
    hydrogen_problem: HydrogenTransportProblem

    non_matching_meshes: bool

    def __init__(
        self,
        heat_problem: HeatTransferProblem,
        hydrogen_problem: HydrogenTransportProblem,
    ) -> None:
        self.heat_problem = heat_problem
        self.hydrogen_problem = hydrogen_problem

        if (
            not self.heat_problem.settings.transient
            or not self.hydrogen_problem.settings.transient
        ):
            raise TypeError(
                "Both the heat and hydrogen problems must be set to transient"
            )

    @property
    def heat_problem(self):
        return self._heat_problem

    @heat_problem.setter
    def heat_problem(self, value):
        if not isinstance(value, HeatTransferProblem):
            raise TypeError("heat_problem must be a festim.HeatTransferProblem object")
        value.show_progress_bar = False
        self._heat_problem = value

    @property
    def hydrogen_problem(self):
        return self._hydrogen_problem

    @hydrogen_problem.setter
    def hydrogen_problem(self, value):
        if isinstance(
            value,
            HydrogenTransportProblemDiscontinuous
            | HydrogenTransportProblemDiscontinuousChangeVar,
        ):
            raise NotImplementedError(
                "Coupled heat transfer - hydrogen transport simulations with "
                "HydrogenTransportProblemDiscontinuousChangeVar or"
                "HydrogenTransportProblemDiscontinuous"
                "not currently supported"
            )
        elif not isinstance(value, HydrogenTransportProblem):
            raise TypeError(
                "hydrogen_problem must be a festim.HydrogenTransportProblem object"
            )
        self._hydrogen_problem = value

    @property
    def non_matching_meshes(self):
        return self.heat_problem.mesh.mesh != self.hydrogen_problem.mesh.mesh

    def initialise(self):
        # make sure both problems have the same initial time step and final time,
        # use minimal initial value of the two and maximal final time of the two
        min_initial_dt = min(
            self.heat_problem.settings.stepsize.initial_value,
            self.hydrogen_problem.settings.stepsize.initial_value,
        )
        self.heat_problem.settings.stepsize.initial_value = min_initial_dt
        self.hydrogen_problem.settings.stepsize.initial_value = min_initial_dt

        if (
            self.heat_problem.settings.final_time
            != self.hydrogen_problem.settings.final_time
        ):
            raise ValueError(
                "Final time values in the heat transfer and hydrogen transport "
                "model must be the same"
            )

        self.heat_problem.initialise()

        if self.non_matching_meshes:
            V = fem.functionspace(self.hydrogen_problem.mesh.mesh, ("P", 1))
            T_func = fem.Function(V)

            nmm_interpolate(T_func, self.heat_problem.u)

            self.hydrogen_problem.temperature = T_func
        else:
            self.hydrogen_problem.temperature = self.heat_problem.u

        self.hydrogen_problem.initialise()

    def iterate(self):
        self.heat_problem.iterate()

        if self.non_matching_meshes:
            nmm_interpolate(
                self.hydrogen_problem.temperature_fenics, self.heat_problem.u
            )

        self.hydrogen_problem.iterate()

        # use the same time step for both problems, use minimum of the two
        next_dt_value = min(
            float(self.hydrogen_problem.dt), float(self.heat_problem.dt)
        )
        self.heat_problem.dt.value = next_dt_value
        self.hydrogen_problem.dt.value = next_dt_value

    def run(self):
        if self.hydrogen_problem.show_progress_bar:
            self.hydrogen_problem.progress_bar = tqdm.autonotebook.tqdm(
                desc=f"Solving {self.__class__.__name__}",
                total=self.hydrogen_problem.settings.final_time,
                unit_scale=True,
            )

        while self.hydrogen_problem.t.value < self.hydrogen_problem.settings.final_time:
            self.iterate()

        if self.hydrogen_problem.show_progress_bar:
            self.hydrogen_problem.progress_bar.refresh()
            self.hydrogen_problem.progress_bar.close()



from mpi4py import MPI

import dolfinx.mesh
import numpy as np
import pytest

import festim as F


def test_different_surface_ids():
    """Checks that different surface ids are correctly set"""
    my_test_model = F.HydrogenTransportProblem()
    my_test_model.species = [F.Species("H")]
    L = 1e-04
    my_test_model.mesh = F.Mesh1D(np.linspace(0, L, num=3))

    surface_subdomains_ids = [3, 8]
    surface_subdomain_1 = F.SurfaceSubdomain1D(id=surface_subdomains_ids[0], x=0)
    surface_subdomain_2 = F.SurfaceSubdomain1D(id=surface_subdomains_ids[1], x=L)
    volume_subdomain = F.VolumeSubdomain1D(id=1, borders=[0, L], material=None)
    my_test_model.subdomains = [
        surface_subdomain_1,
        surface_subdomain_2,
        volume_subdomain,
    ]

    my_test_model.define_function_spaces()
    my_test_model.define_meshtags_and_measures()

    for surf_id in surface_subdomains_ids:
        assert surf_id in np.array(my_test_model.facet_meshtags.values)


def test_different_volume_ids():
    my_model = F.HydrogenTransportProblem()
    my_model.mesh = F.Mesh1D(vertices=np.linspace(0, 10, 11))

    vol_subdom_ids = [2, 16, 7]
    vol_subdomain_1 = F.VolumeSubdomain1D(
        id=vol_subdom_ids[0], borders=[0, 2], material=None
    )
    vol_subdomain_2 = F.VolumeSubdomain1D(
        id=vol_subdom_ids[1], borders=[2, 6], material=None
    )
    vol_subdomain_3 = F.VolumeSubdomain1D(
        id=vol_subdom_ids[2], borders=[6, 10], material=None
    )
    my_model.subdomains = [vol_subdomain_1, vol_subdomain_2, vol_subdomain_3]

    my_model.define_meshtags_and_measures()

    for vol_id in vol_subdom_ids:
        assert vol_id in np.array(my_model.volume_meshtags.values)


def test_non_matching_volume_borders():
    """Checks that non-matching borders raise an error"""
    mesh = F.Mesh1D(vertices=np.linspace(0, 5, 6))
    vol_subdomain_1 = F.VolumeSubdomain1D(id=1, borders=[0, 2], material=None)
    vol_subdomain_2 = F.VolumeSubdomain1D(id=1, borders=[3, 5], material=None)
    subdomains = [vol_subdomain_1, vol_subdomain_2]

    with pytest.raises(ValueError, match="Subdomain borders don't match to each other"):
        mesh.check_borders(subdomains)


def test_matching_volume_borders_non_ascending_order():
    """Checks that subdomain placed in non ascending order still passes"""
    mesh = F.Mesh1D(vertices=np.linspace(0, 8, 9))
    vol_subdomain_1 = F.VolumeSubdomain1D(id=1, borders=[0, 2], material=None)
    vol_subdomain_2 = F.VolumeSubdomain1D(id=2, borders=[4, 8], material=None)
    vol_subdomain_3 = F.VolumeSubdomain1D(id=3, borders=[2, 4], material=None)
    subdomains = [vol_subdomain_1, vol_subdomain_2, vol_subdomain_3]

    mesh.check_borders(subdomains)


def test_borders_out_of_domain():
    """Checks that borders outside of the domain raise an error"""
    mesh = F.Mesh1D(vertices=np.linspace(0, 2))
    subdomains = [F.VolumeSubdomain1D(id=1, borders=[1, 15], material=None)]
    with pytest.raises(ValueError, match="borders dont match domain borders"):
        mesh.check_borders(subdomains)


def test_borders_inside_domain():
    """Checks that borders inside of the domain raise an error"""
    mesh = F.Mesh1D(vertices=np.linspace(0, 20))
    subdomains = [F.VolumeSubdomain1D(id=1, borders=[1, 6], material=None)]
    with pytest.raises(ValueError, match="borders dont match domain borders"):
        mesh.check_borders(subdomains)


def test_raise_error_with_no_volume_subdomain():
    """Checks that error is rasied when no volume subdomain is defined"""
    mesh = F.Mesh1D(vertices=np.linspace(0, 20))

    with pytest.raises(ValueError, match="No volume subdomains defined"):
        mesh.check_borders([])


@pytest.mark.parametrize("input", [1, 2, 3, 4])
def test_find_volume_from_int(input):
    """test that the correct volume is returned when input is an int"""

    vol_1 = F.VolumeSubdomain1D(id=1, borders=[0, 1], material=None)
    vol_2 = F.VolumeSubdomain1D(id=2, borders=[1, 2], material=None)
    vol_3 = F.VolumeSubdomain1D(id=3, borders=[2, 3], material=None)
    vol_4 = F.VolumeSubdomain1D(id=4, borders=[3, 4], material=None)

    volumes = [vol_1, vol_2, vol_3, vol_4]

    assert F.find_volume_from_id(input, volumes) == volumes[input - 1]


def test_ValueError_raised_when_id_not_found_in_volumes_subdomains():
    """test that a ValueError is raised when an id is not found in the list of volume subdomains"""

    volumes = [F.VolumeSubdomain1D(id=1, borders=[0, 1], material=None)]

    with pytest.raises(ValueError, match="id 5 not found in list of volumes"):
        F.find_volume_from_id(5, volumes)


def test_ValueError_rasied_when_volume_ids_are_not_unique():
    """Checks"""
    my_test_model = F.HydrogenTransportProblem(
        mesh=F.Mesh1D(np.linspace(0, 2, num=10)), species=[F.Species("H")]
    )

    vol_1 = F.VolumeSubdomain1D(id=1, borders=[0, 1], material=None)
    vol_2 = F.VolumeSubdomain1D(id=1, borders=[1, 2], material=None)
    my_test_model.subdomains = [
        vol_1,
        vol_2,
    ]

    my_test_model.define_function_spaces()

    with pytest.raises(ValueError, match="Volume ids are not unique"):
        my_test_model.define_meshtags_and_measures()


def test_ValueError_raised_when_id_not_found_in_surface_subdomains():
    """test that a ValueError is raised when an id is not found in the list of surface subdomains"""

    surfaces = [F.SurfaceSubdomain(id=1)]

    with pytest.raises(ValueError, match="id 3 not found in list of surfaces"):
        F.find_surface_from_id(3, surfaces)


@pytest.mark.parametrize("input_id, output_index", [(1, 2), (4, 0), (7, 1), (9, 3)])
def test_find_surface_from_id(input_id, output_index):
    """test that the correct surface is returned when input is an int"""

    surf_1 = F.SurfaceSubdomain(id=7)
    surf_2 = F.SurfaceSubdomain1D(id=4, x=0)
    surf_3 = F.SurfaceSubdomain1D(id=9, x=1)
    surf_4 = F.SurfaceSubdomain(id=1)

    surfaces = [surf_2, surf_1, surf_4, surf_3]

    assert F.find_surface_from_id(input_id, surfaces) == surfaces[output_index]


def test_volume_subdomain_properties():
    """Tests that the volume subdomain property obtains the correct
    subdomains from the the model subdomains list"""

    my_model = F.HydrogenTransportProblem()
    my_model.subdomains = [
        F.SurfaceSubdomain(id=7),
        F.SurfaceSubdomain1D(id=4, x=0),
        F.SurfaceSubdomain(id=2),
        F.VolumeSubdomain(id=1, material=None),
        F.VolumeSubdomain1D(id=9, borders=[0, 1], material=None),
    ]

    assert len(my_model.volume_subdomains) == 2
    for subdomain in my_model.volume_subdomains:
        assert isinstance(subdomain, F.VolumeSubdomain)

    assert len(my_model.surface_subdomains) == 3
    for subdomain in my_model.surface_subdomains:
        assert isinstance(subdomain, F.SurfaceSubdomain)


def test_subdomain_with_custom_locator():
    mesh = mesh2d
    subdomain1 = F.SurfaceSubdomain(id=1, locator=lambda x: np.isclose(x[0], 0))
    subdomain2 = F.SurfaceSubdomain(id=1, locator=lambda x: np.full(x.shape[1], True))

    for subdomain in [subdomain1, subdomain2]:
        # Check that the subdomain can be located
        entities = subdomain.locate_boundary_facet_indices(mesh)
        assert len(entities) > 0


mesh1d = dolfinx.mesh.create_unit_interval(MPI.COMM_WORLD, nx=1)
mesh2d = dolfinx.mesh.create_unit_square(MPI.COMM_WORLD, nx=1, ny=1)
mesh3d = dolfinx.mesh.create_unit_cube(MPI.COMM_WORLD, nx=1, ny=1, nz=1)



from mpi4py import MPI

import dolfinx
import numpy as np

from festim.exports.volume_quantity import VolumeQuantity


class MinimumVolume(VolumeQuantity):
    """Computes the minimum value of a field in a given volume

    Args:
        field (festim.Species): species for which the minimum volume is computed
        volume (festim.VolumeSubdomain): volume subdomain
        filename (str, optional): name of the file to which the minimum volume is exported

    Attributes:
        see `festim.VolumeQuantity`
    """

    @property
    def title(self):
        return f"Minimum {self.field.name} volume {self.volume.id}"

    def compute(self):
        """
        Computes the minimum value of solution function within the defined volume
        subdomain, and appends it to the data list
        """
        solution = self.field.post_processing_solution
        entities = self.volume_meshtags.find(self.volume.id)

        if isinstance(solution, dolfinx.fem.Function):
            V = solution.function_space
        else:
            V = self.field.sub_function_space
        mesh = V.mesh
        mesh.topology.create_connectivity(mesh.topology.dim, mesh.topology.dim)
        dofs = dolfinx.fem.locate_dofs_topological(
            V=V, entity_dim=mesh.topology.dim, entities=entities
        )

        self.value = mesh.comm.allreduce(np.min(solution.x.array[dofs]), op=MPI.MIN)
        self.data.append(self.value)



from enum import Enum

import dolfinx
import numpy as np
import ufl
from dolfinx.mesh import Mesh as dolfinx_Mesh
from dolfinx.mesh import meshtags

__all__ = ["CoordinateSystem", "Mesh"]


class CoordinateSystem(Enum):
    CARTESIAN = 10
    CYLINDRICAL = 20
    SPHERICAL = 30

    @classmethod
    def from_string(cls, s: str):
        """Can be removed with Python 3.11+."""
        s = s.lower()
        if s == "cartesian":
            return cls.CARTESIAN
        elif s == "cylindrical":
            return cls.CYLINDRICAL
        elif s == "spherical":
            return cls.SPHERICAL
        else:
            raise ValueError(
                "coordinate_system must be one of 'cartesian', 'cylindrical', or 'spherical'"
            )

    def __str__(self):
        if self == CoordinateSystem.CARTESIAN:
            return "cartesian"
        elif self == CoordinateSystem.CYLINDRICAL:
            return "cylindrical"
        elif self == CoordinateSystem.SPHERICAL:
            return "spherical"
        else:
            raise ValueError("Invalid CoordinateSystem value")


class Mesh:
    """
    Mesh class

    Args:
        mesh: The mesh. Defaults to None.
        coordinate_system: the coordinate system of the mesh ("cartesian",
            "cylindrical", "spherical"). Defaults to "cartesian".

    Attributes:
        mesh The mesh
        vdim: the dimension of the mesh cells
        fdim: the dimension of the mesh facets
        n: Symbolic representation of the vector normal to the facets
            of the mesh.
    """

    mesh: dolfinx.mesh.Mesh
    _coordinate_system: CoordinateSystem

    vdim: int
    fdim: int
    n: ufl.FacetNormal

    def __init__(
        self,
        mesh: dolfinx_Mesh | None = None,
        coordinate_system: str | CoordinateSystem = CoordinateSystem.CARTESIAN,
    ):
        self.mesh = mesh
        if self._mesh is not None:
            # create cell to facet connectivity
            self._mesh.topology.create_connectivity(
                self._mesh.topology.dim, self._mesh.topology.dim - 1
            )

            # create facet to cell connectivity
            self._mesh.topology.create_connectivity(
                self._mesh.topology.dim - 1, self._mesh.topology.dim
            )
            self.coordinate_system = coordinate_system

        self.check_mesh_dim_coords()

    @property
    def mesh(self):
        return self._mesh

    @mesh.setter
    def mesh(self, value):
        if isinstance(value, dolfinx.mesh.Mesh):
            self._mesh = value
        else:
            raise TypeError("Mesh must be of type dolfinx.mesh.Mesh")

    @property
    def coordinate_system(self):
        return self._coordinate_system

    @coordinate_system.setter
    def coordinate_system(self, value):
        if isinstance(value, CoordinateSystem):
            self._coordinate_system = value
        elif isinstance(value, str):
            self._coordinate_system = CoordinateSystem.from_string(value)
        else:
            raise ValueError(
                "coordinate_system must be of type str or CoordinateSystem"
            )

    @property
    def vdim(self):
        if self._mesh is None:
            raise RuntimeError("Mesh is not defined")
        return self._mesh.topology.dim

    @property
    def fdim(self):
        if self._mesh is None:
            raise RuntimeError("Mesh is not defined")
        return self._mesh.topology.dim - 1

    @property
    def n(self):
        if self._mesh is None:
            raise RuntimeError("Mesh is not defined")
        return ufl.FacetNormal(self._mesh)

    def define_meshtags(self, surface_subdomains, volume_subdomains, interfaces=None):
        """Defines the facet and volume meshtags of the mesh

        Args:
            surface_subdomains (list of festim.SufaceSubdomains): the surface subdomains of the model
            volume_subdomains (list of festim.VolumeSubdomains): the volume subdomains of the model
            interfaces (dict, optional): the interfaces between volume
                subdomains {int: [VolumeSubdomain, VolumeSubdomain]}. Defaults to None.

        Returns:
            dolfinx.mesh.MeshTags: the facet meshtags
            dolfinx.mesh.MeshTags: the volume meshtags
        """
        # find all cells in domain and mark them as 0
        num_cells = self._mesh.topology.index_map(self.vdim).size_local
        mesh_cell_indices = np.arange(num_cells, dtype=np.int32)
        tags_volumes = np.full(num_cells, 0, dtype=np.int32)

        # find all facets in domain and mark them as 0
        num_facets = self._mesh.topology.index_map(self.fdim).size_local
        mesh_facet_indices = np.arange(num_facets, dtype=np.int32)
        tags_facets = np.full(num_facets, 0, dtype=np.int32)

        for surf in surface_subdomains:
            try:
                # find all facets in subdomain and mark them as surf.id
                entities = surf.locate_boundary_facet_indices(self._mesh)
                tags_facets[entities] = surf.id
            except ValueError:
                if len(surface_subdomains) > 1:
                    raise ValueError(
                        "Surface subdomain must have a locator attribute if"
                        " several subdomains are defined"
                    )
                self.mesh.topology.create_connectivity(self.fdim, self.fdim + 1)
                rentities = dolfinx.mesh.exterior_facet_indices(self.mesh.topology)

                tags_facets[rentities] = surf.id

        for vol in volume_subdomains:
            try:
                # find all cells in subdomain and mark them as vol.id
                entities = vol.locate_subdomain_entities(self._mesh)
                tags_volumes[entities] = vol.id
            except ValueError:
                if len(volume_subdomains) > 1:
                    raise ValueError(
                        "Volume subdomain must have a locator if"
                        " several subdomains are defined"
                    )
                tags_volumes[:] = vol.id

        volume_meshtags = meshtags(
            self._mesh, self.vdim, mesh_cell_indices, tags_volumes
        )

        # tag interfaces
        interfaces = interfaces or {}  # if interfaces is None, set it to empty dict
        for interface in interfaces:
            (domain_0, domain_1) = interface.subdomains
            all_0_facets = dolfinx.mesh.compute_incident_entities(
                self._mesh.topology,
                volume_meshtags.find(domain_0.id),
                self.vdim,
                self.fdim,
            )
            all_1_facets = dolfinx.mesh.compute_incident_entities(
                self._mesh.topology,
                volume_meshtags.find(domain_1.id),
                self.vdim,
                self.fdim,
            )
            interface_entities = np.intersect1d(all_0_facets, all_1_facets)
            tags_facets[interface_entities] = interface.id

        facet_meshtags = meshtags(
            self._mesh, self.fdim, mesh_facet_indices, tags_facets
        )

        return facet_meshtags, volume_meshtags

    def check_mesh_dim_coords(self):
        """Checks if the used coordinates can be applied for geometry with the specified
        dimensions"""

        if self.coordinate_system == CoordinateSystem.SPHERICAL and self.vdim != 1:
            raise AttributeError(
                "spherical coordinates can be used for one-dimensional domains only"
            )
        if self.coordinate_system == CoordinateSystem.CYLINDRICAL and self.vdim == 3:
            raise AttributeError(
                "cylindrical coordinates cannot be used for 3D domains"
            )



import numpy as np


class Stepsize:
    """
    A class for evaluating the stepsize of transient simulations.

    Args:
        initial_value (float, int): initial stepsize.
        growth_factor (float, optional): factor by which the stepsize is
            increased when adapting
        cutback_factor (float, optional): factor by which the stepsize is
            decreased when adapting
        target_nb_iterations (int, optional): number of Newton iterations
            over (resp. under) which the stepsize is increased
            (resp. decreased)
        max_stepsize (float or callable, optional): Maximum stepsize.
            If callable, has to be a function of `t`. Defaults to None.
        milestones (list, optional): list of times by which the simulation must
            pass. Defaults to an empty list.


    Attributes:
        initial_value (float, int): initial stepsize.
        growth_factor (float): factor by which the stepsize is
            increased when adapting
        cutback_factor (float): factor by which the stepsize is
            decreased when adapting
        target_nb_iterations (int): number of Newton iterations
            over (resp. under) which the stepsize is increased
            (resp. decreased)
        adaptive (bool): True if the stepsize is adaptive, False otherwise.
        max_stepsize (float, callable): Maximum stepsize.
        milestones (list): list of times by which the simulation must
            pass.
    """

    def __init__(
        self,
        initial_value,
        growth_factor=None,
        cutback_factor=None,
        target_nb_iterations=None,
        max_stepsize=None,
        milestones=None,
    ) -> None:
        self.initial_value = initial_value
        self.growth_factor = growth_factor
        self.cutback_factor = cutback_factor
        self.target_nb_iterations = target_nb_iterations
        self.max_stepsize = max_stepsize
        self.milestones = milestones or []

        # TODO should this class hold the dt object used in the formulation

    @property
    def milestones(self):
        return self._milestones

    @milestones.setter
    def milestones(self, value):
        if value:
            self._milestones = sorted(value)
        else:
            self._milestones = value

    @property
    def adaptive(self):
        return self.growth_factor or self.cutback_factor or self.target_nb_iterations

    @property
    def growth_factor(self):
        return self._growth_factor

    @growth_factor.setter
    def growth_factor(self, value):
        if value is not None:
            if value < 1:
                raise ValueError("growth factor should be greater than one")

        self._growth_factor = value

    @property
    def cutback_factor(self):
        return self._cutback_factor

    @cutback_factor.setter
    def cutback_factor(self, value):
        if value is not None:
            if value > 1:
                raise ValueError("cutback factor should be smaller than one")

        self._cutback_factor = value

    @property
    def max_stepsize(self):
        return self._max_stepsize

    @max_stepsize.setter
    def max_stepsize(self, value):
        if isinstance(value, float):
            if value < self.initial_value:
                raise ValueError(
                    "maximum stepsize cannot be less than initial stepsize"
                )

        self._max_stepsize = value

    def get_max_stepsize(self, t):
        """
        Returns the maximum stepsize at time t.

        Args:
            t (float): the current time

        Returns:
            float or None: the maximum stepsize at time t
        """
        if callable(self._max_stepsize):
            return self._max_stepsize(t)
        return self._max_stepsize

    def modify_value(self, value, nb_iterations, t=None):
        if not self.is_adapt(t):
            return value

        if nb_iterations < self.target_nb_iterations:
            updated_value = value * self.growth_factor
        elif nb_iterations > self.target_nb_iterations:
            updated_value = value * self.cutback_factor
        else:
            updated_value = value

        if max_step := self.get_max_stepsize(t):
            if updated_value > max_step:
                updated_value = max_step

        next_milestone = self.next_milestone(t)
        if next_milestone is not None:
            time_to_milestone = next_milestone - t
            if updated_value > time_to_milestone and not np.isclose(
                t, next_milestone, atol=0
            ):
                updated_value = time_to_milestone

        return updated_value

    def is_adapt(self, t):
        """
        Methods that defines if the stepsize should be
        adapted or not

        Args:
            t (float): the current time

        Returns:
            bool: True if needs to adapt, False otherwise.
        """
        return True

    def next_milestone(self, current_time: float):
        """Returns the next milestone that the simulation must pass.
        Returns None if there are no more milestones.

        Args:
            current_time (float): current time.

        Returns:
            float: next milestone.
        """
        if self.milestones is None:
            return None
        for milestone in self.milestones:
            if current_time < milestone:
                return milestone
        return None



import numpy as np
import pytest

import festim as F


def concentration_A_exact(t, c_A_0, k, p):
    """Analytical solution for the concentration of species A in a reaction A + B <-> C + D
    assuming [A]_0 = [B]_0 and [C]_0 = [D]_0 = 0
    can be obtained by solving
    d[A]/dt = -k[A][B] + p[C][D]
    where [A] = [B] and [C] = [D] = 0 at t=0
    Moreover, [C] = [D] = [A]_0 - [A]
    The equation then becomes
    d[A]/dt = -k[A]^2 + p([A]_0 - [A])^2

    Args:
        t (float or ndarray): time
        c_A_0 (float): initial concentration of A
        k (float): forward reaction rate
        p (float): backward reaction rate

    Returns:
        ndarray: the concentration of A at a given time
    """

    A = 1 - p / k
    B = 2 * p / k * c_A_0
    C = -p / k * c_A_0**2
    roots = np.roots([A, B, C])

    if len(roots) == 1:
        a = roots[0]
        b = roots[0]
    else:
        a, b = roots[0], roots[1]
    F = ((c_A_0 - a) / (c_A_0 - b)) * np.exp(-(a - b) * (k - p) * t)
    return (a - b * F) / (1 - F)


def model_test_reaction(stepsize=1, k=350e-4, p=120e-4, c_A_0=1):
    """Creates a festim model with a single reaction and runs it.
    The reaction is A + B <-> C + D

    Args:
        stepsize (float): the stepsize
        k (float): the forward reaction rate
        p (float): the backward reaction rate
        c_A_0 (float): the initial concentration of A and B

    Returns:
        festim.HydrogenTransportProblem: the model
    """

    my_model = F.HydrogenTransportProblem()

    # -------- Mesh --------- #

    L = 1
    vertices = np.linspace(0, L, num=10)
    my_model.mesh = F.Mesh1D(vertices)

    # -------- Materials and subdomains --------- #

    mat = F.Material(D_0=1, E_D=0, name="mat")

    my_subdomain = F.VolumeSubdomain1D(id=1, borders=[0, L], material=mat)
    left_surface = F.SurfaceSubdomain1D(id=1, x=0)
    right_surface = F.SurfaceSubdomain1D(id=2, x=L)

    my_model.subdomains = [
        my_subdomain,
        left_surface,
        right_surface,
    ]

    # -------- Hydrogen species and reactions --------- #

    species_A = F.Species("A")
    species_B = F.Species("B")
    species_C = F.Species("C")
    species_D = F.Species("D")

    my_model.species = [
        species_A,
        species_B,
        species_C,
        species_D,
    ]

    my_model.reactions = [
        F.Reaction(
            k_0=k,
            E_k=0,
            p_0=p,
            E_p=0,
            reactant=[species_A, species_B],
            product=[species_C, species_D],
            volume=my_subdomain,
        ),
    ]

    # -------- Temperature --------- #

    my_model.temperature = 400

    # -------- Boundary conditions --------- #

    my_model.boundary_conditions = []

    # -------- Initial conditions --------- #

    my_model.initial_conditions = [
        F.InitialConcentration(value=c_A_0, species=species_A, volume=my_subdomain),
        F.InitialConcentration(value=c_A_0, species=species_B, volume=my_subdomain),
    ]

    # -------- Exports --------- #

    total_A = F.TotalVolume(species_A, my_subdomain)
    total_B = F.TotalVolume(species_B, my_subdomain)
    total_C = F.TotalVolume(species_C, my_subdomain)
    total_D = F.TotalVolume(species_D, my_subdomain)

    my_model.exports = [total_A, total_B, total_C, total_D]

    # -------- Settings --------- #

    my_model.settings = F.Settings(
        atol=1e-10, rtol=1e-10, max_iterations=30, final_time=200
    )

    my_model.settings.stepsize = F.Stepsize(initial_value=stepsize)

    # -------- Run --------- #

    my_model.initialise()

    my_model.run()

    return my_model


def compute_error(model):
    """Computes the relative L2 error norm between the analytical solution and the numerical solution

    Args:
        model (F.HydrogenTransportModel): the numerical festim model

    Returns:
        float: the relative L2 error norm
    """
    L = model.mesh.vertices[-1]
    k = model.reactions[0].k_0
    p = model.reactions[0].p_0
    c_A_0 = model.initial_conditions[0].value
    total_A = model.exports[0]
    times = np.array(total_A.t)
    data_total_A = total_A.data
    data_total_A = np.array(data_total_A)
    concentration_A_analytical = concentration_A_exact(t=times, c_A_0=c_A_0, k=k, p=p)

    # relative l2 error norm
    error = np.sqrt(
        np.sum((concentration_A_analytical - data_total_A / L) ** 2)
    ) / np.sqrt(np.sum(concentration_A_analytical**2))

    return error


@pytest.mark.parametrize("k, p, c_A_0", [(350e-4, 120e-4, 3), (200e-4, 100e-4, 2)])
def test_reaction(k, p, c_A_0):
    """Test the reaction A + B <-> C + D with a festim model and compare the results with the analytical solution"""
    model = model_test_reaction(stepsize=0.5, k=k, p=p, c_A_0=c_A_0)
    error = compute_error(model)
    assert error < 1e-2



import warnings
from pathlib import Path

from dolfinx import fem, io

from festim.species import Species
from festim.subdomain.volume_subdomain import VolumeSubdomain


class ExportBaseClass:
    """Export functions to VTX file

    Args:
        filename: The name of the output file
        times: if provided, the field will be exported at these timesteps. Otherwise
            exports at all timesteps. Defaults to None.

    Attributes:
        filename: The name of the output file
        times: if provided, the field will be exported at these timesteps. Otherwise
            exports at all timesteps. Defaults to None.
    """

    _filename: Path | str
    writer: io.VTXWriter

    def __init__(
        self,
        filename: str | Path,
        ext: str,
        times: list[float] | list[int] | None | None = None,
    ):
        name = Path(filename)
        if name.suffix != ext:
            warnings.warn(
                f"Filename {filename} does not have {ext} extension, adding it."
            )
            name = name.with_suffix(ext)

        self._filename = name
        if times:
            self.times = sorted(times)
        else:
            self.times = times

    @property
    def filename(self):
        return self._filename


class VTXTemperatureExport(ExportBaseClass):
    """Export temperature field functions to VTX file

    Args:
        filename: The name of the output file
        times: if provided, the field will be exported at these timesteps. Otherwise
            exports at all timesteps. Defaults to None.

    Attributes:
        filename: The name of the output file
        times: if provided, the field will be exported at these timesteps. Otherwise
            exports at all timesteps. Defaults to None.
        writer: The VTXWriter object used to write the file
    """

    writer: io.VTXWriter

    def __init__(
        self,
        filename: str | Path,
        times: list[float] | list[int] | None | None = None,
    ):
        super().__init__(filename, ".bp", times)


class VTXSpeciesExport(ExportBaseClass):
    """Export species field functions to VTX file

    Args:
        filename: The name of the output file
        field: Set of species to export
        subdomain: A field can be defined on multiple domains. This arguments specifies
            what subdomains we export on. If `None` we export on all domains.
        checkpoint: If True, the export will be a checkpoint file using adios4dolfinx
            and won't be readable by ParaView. Default is False.
        times: if provided, the field will be exported at these timesteps. Otherwise
            exports at all timesteps. Defaults to None.

    Attributes:
        filename: The name of the output file
        field: Set of species to export
        times: if provided, the field will be exported at these timesteps. Otherwise
            exports at all timesteps. Defaults to None.
        writer: The VTXWriter object used to write the file
    """

    _subdomain: VolumeSubdomain
    _checkpoint: bool
    writer: io.VTXWriter

    def __init__(
        self,
        filename: str | Path,
        field: Species | list[Species],
        subdomain: VolumeSubdomain = None,
        checkpoint: bool = False,
        times: list[float] | list[int] | None | None = None,
    ):
        super().__init__(filename, ".bp", times)
        self.field = field
        self._subdomain = subdomain
        self._checkpoint = checkpoint

    @property
    def field(self) -> list[Species]:
        return self._field

    @field.setter
    def field(self, value: Species | list[Species]):
        """
        Update the field to export.

        Args:
            value: The species to export

        Raises:
            TypeError: If input field is not a Species or a list of Species

        Note:
            This also creates a new writer with the updated field.
        """
        # check that all elements of list are festim.Species
        if isinstance(value, list):
            for element in value:
                if not isinstance(element, Species | str):
                    raise TypeError(
                        "field must be of type festim.Species or a list of "
                        "festim.Species or str"
                    )
            val = value
        elif isinstance(value, Species):
            val = [value]
        else:
            raise TypeError(
                "field must be of type festim.Species or a list of festim.Species or "
                "str",
                f"got {type(value)}.",
            )
        self._field = val

    def get_functions(self) -> list[fem.Function]:
        """
        Returns list of species for a given subdomain. If using legacy mode, return the
        whole species.
        """

        legacy_output: bool = False
        for field in self._field:
            if field.legacy:
                legacy_output = True
                break
        if legacy_output:
            return [field.post_processing_solution for field in self._field]
        else:
            if self._subdomain is None:
                raise ValueError("Subdomain must be specified")
            else:
                outfiles = []
                for field in self._field:
                    if self._subdomain in field.subdomains:
                        outfiles.append(
                            field.subdomain_to_post_processing_solution[self._subdomain]
                        )
                return outfiles



from mpi4py import MPI

import basix
import dolfinx
import dolfinx.fem.petsc
import numpy as np

import festim as F


# ---------------- Generate a mesh ----------------
def generate_mesh(n=20):
    def bottom_boundary(x):
        return np.isclose(x[1], 0.0)

    def top_boundary(x):
        return np.isclose(x[1], 1.0)

    def bottom_left_boundary(x):
        return np.logical_and(np.isclose(x[0], 0.0), x[1] <= 0.5 + 1e-14)

    def half(x):
        return x[1] <= 0.5 + 1e-14

    mesh = dolfinx.mesh.create_unit_square(
        MPI.COMM_WORLD, n, n, dolfinx.mesh.CellType.triangle
    )

    # Split domain in half and set an interface tag of 5
    gdim = mesh.geometry.dim
    tdim = mesh.topology.dim
    fdim = tdim - 1
    top_facets = dolfinx.mesh.locate_entities_boundary(mesh, fdim, top_boundary)
    bottom_facets = dolfinx.mesh.locate_entities_boundary(mesh, fdim, bottom_boundary)
    bot_left_facets = dolfinx.mesh.locate_entities_boundary(
        mesh, fdim, bottom_left_boundary
    )
    num_facets_local = (
        mesh.topology.index_map(fdim).size_local
        + mesh.topology.index_map(fdim).num_ghosts
    )
    facets = np.arange(num_facets_local, dtype=np.int32)
    values = np.full_like(facets, 0, dtype=np.int32)
    values[top_facets] = 1
    values[bottom_facets] = 2
    values[bot_left_facets] = 3

    bottom_cells = dolfinx.mesh.locate_entities(mesh, tdim, half)
    num_cells_local = (
        mesh.topology.index_map(tdim).size_local
        + mesh.topology.index_map(tdim).num_ghosts
    )
    cells = np.full(num_cells_local, 5, dtype=np.int32)
    cells[bottom_cells] = 4
    ct = dolfinx.mesh.meshtags(
        mesh, tdim, np.arange(num_cells_local, dtype=np.int32), cells
    )
    all_b_facets = dolfinx.mesh.compute_incident_entities(
        mesh.topology, ct.find(4), tdim, fdim
    )
    all_t_facets = dolfinx.mesh.compute_incident_entities(
        mesh.topology, ct.find(5), tdim, fdim
    )
    interface = np.intersect1d(all_b_facets, all_t_facets)
    values[interface] = 6

    mt = dolfinx.mesh.meshtags(mesh, mesh.topology.dim - 1, facets, values)
    return mesh, mt, ct


mesh, mt, ct = generate_mesh()


# create velocity field
def create_velocity_field():
    def velocity_func(x):
        values = np.zeros((2, x.shape[1]))  # Initialize with zeros
        scalar_value = -500 * x[1] * (x[1] - 0.5)  # Compute the scalar function
        values[0] = scalar_value  # Assign to first component
        values[1] = 0  # Second component remains zero
        return values

    mesh2, mt2, ct2 = generate_mesh(n=10)
    submesh, cell_map, v_map = dolfinx.mesh.create_submesh(mesh2, ct2.dim, ct2.find(4))[
        0:3
    ]
    v_cg = basix.ufl.element(
        "Lagrange", submesh.topology.cell_name(), 2, shape=(submesh.geometry.dim,)
    )
    V_velocity = dolfinx.fem.functionspace(submesh, v_cg)
    u = dolfinx.fem.Function(V_velocity)
    u.interpolate(velocity_func)
    return u


u = create_velocity_field()

writer = dolfinx.io.VTXWriter(MPI.COMM_WORLD, "velocity.bp", u, engine="BP5")
writer.write(t=0)

my_model = F.HydrogenTransportProblemDiscontinuous()
my_model.mesh = F.Mesh(mesh)
my_model.volume_meshtags = ct
my_model.facet_meshtags = mt

material_bottom = F.Material(D_0=1, E_D=0 * 0.1)
material_top = F.Material(D_0=1, E_D=0 * 0.1)

material_bottom.K_S_0 = 2.0
material_bottom.E_K_S = 0
material_top.K_S_0 = 600
material_top.E_K_S = 0

top_domain = F.VolumeSubdomain(5, material=material_top)
bottom_domain = F.VolumeSubdomain(4, material=material_bottom)

top_surface = F.SurfaceSubdomain(id=1)
bottom_surface = F.SurfaceSubdomain(id=2)
inlet_surf = F.SurfaceSubdomain(id=3)
my_model.subdomains = [
    bottom_domain,
    top_domain,
    top_surface,
    bottom_surface,
    inlet_surf,
]

# we should be able to automate this
my_model.interfaces = [F.Interface(6, (bottom_domain, top_domain))]
my_model.surface_to_volume = {
    top_surface: top_domain,
    bottom_surface: bottom_domain,
    inlet_surf: bottom_domain,
}

H = F.Species("H", mobile=True)

my_model.species = [H]

for species in my_model.species:
    species.subdomains = [bottom_domain, top_domain]


my_model.boundary_conditions = [
    F.FixedConcentrationBC(top_surface, value=0.0, species=H),
    F.FixedConcentrationBC(inlet_surf, value=1.0, species=H),
    F.FixedConcentrationBC(bottom_surface, value=0.0, species=H),
]

my_model.advection_terms = [
    F.AdvectionTerm(velocity=u, species=[H], subdomain=bottom_domain)
]


my_model.temperature = 500.0

my_model.settings = F.Settings(atol=1e-10, rtol=1e-10, transient=False)

my_model.exports = [
    F.VTXSpeciesExport(f"u_{subdomain.id}.bp", field=H, subdomain=subdomain)
    for subdomain in my_model.volume_subdomains
]

my_model.initialise()
my_model.run()



from .dirichlet_bc import DirichletBCBase, FixedConcentrationBC, FixedTemperatureBC
from .flux_bc import FluxBCBase, HeatFluxBC, ParticleFluxBC
from .henrys_bc import HenrysBC
from .sieverts_bc import SievertsBC
from .surface_reaction import SurfaceReactionBC, SurfaceReactionBCpartial

__all__ = [
    "DirichletBCBase",
    "FixedConcentrationBC",
    "FixedTemperatureBC",
    "FluxBCBase",
    "HeatFluxBC",
    "HenrysBC",
    "ParticleFluxBC",
    "SievertsBC",
    "SurfaceReactionBC",
    "SurfaceReactionBCpartial",
]



import numpy as np
import pytest

import festim as F

test_mesh = F.Mesh1D(vertices=np.linspace(0, 1, 100))
test_mat = F.Material(D_0=1, E_D=0, thermal_conductivity=1, heat_capacity=1, density=1)
test_subdomains = [F.VolumeSubdomain1D(id=1, borders=[0, 1], material=test_mat)]
test_H = F.Species("H")


@pytest.mark.parametrize(
    "object",
    ["coucou", 1, 1.0],
)
def test_error_raised_when_wrong_hydrogen_problem_given(object):
    """Test TypeError is raised when an object that isnt a
    festim.HydrogenTransportProblem object is given to hydrogen_problem"""

    test_heat_problem = F.HeatTransferProblem()

    with pytest.raises(
        TypeError,
        match="hydrogen_problem must be a festim.HydrogenTransportProblem object",
    ):
        F.CoupledTransientHeatTransferHydrogenTransport(
            heat_problem=test_heat_problem, hydrogen_problem=object
        )


@pytest.mark.parametrize(
    "object",
    [
        F.HydrogenTransportProblemDiscontinuous(),
        F.HydrogenTransportProblemDiscontinuousChangeVar(),
    ],
)
def test_error_raised_when_wrong_type_hydrogen_problem_given(object):
    """Test TypeError is raised when an object that isnt a
    festim.HydrogenTransportProblem object is given to hydrogen_problem"""

    test_heat_problem = F.HeatTransferProblem()

    with pytest.raises(
        NotImplementedError,
        match="Coupled heat transfer - hydrogen transport simulations with "
        "HydrogenTransportProblemDiscontinuousChangeVar or"
        "HydrogenTransportProblemDiscontinuous"
        "not currently supported",
    ):
        F.CoupledTransientHeatTransferHydrogenTransport(
            heat_problem=test_heat_problem, hydrogen_problem=object
        )


@pytest.mark.parametrize(
    "object",
    ["coucou", 1, 1.0],
)
def test_error_raised_when_wrong_heat_problem_given(object):
    """Test TypeError is raised when an object that isnt a
    festim.HeatTransferProblem object is given to heat_problem"""

    test_hydrogen_problem = F.HydrogenTransportProblem()

    with pytest.raises(
        TypeError,
        match="heat_problem must be a festim.HeatTransferProblem object",
    ):
        F.CoupledTransientHeatTransferHydrogenTransport(
            heat_problem=object, hydrogen_problem=test_hydrogen_problem
        )


def test_initial_dt_values_are_the_same():
    """Test that the smallest of the stepsize intial_value values given is used in both
    the heat_problem and the hydrogen_problem"""

    test_heat_problem = F.HeatTransferProblem(
        mesh=test_mesh,
        subdomains=test_subdomains,
        settings=F.Settings(
            atol=1, rtol=0.9999, transient=True, stepsize=0.5, final_time=5
        ),
    )

    test_hydrogen_problem = F.HydrogenTransportProblem(
        mesh=test_mesh,
        subdomains=test_subdomains,
        species=[test_H],
        settings=F.Settings(
            atol=1, rtol=0.9999, transient=True, stepsize=1.5, final_time=5
        ),
    )

    test_coupled_problem = F.CoupledTransientHeatTransferHydrogenTransport(
        heat_problem=test_heat_problem,
        hydrogen_problem=test_hydrogen_problem,
    )

    test_coupled_problem.initialise()

    assert np.isclose(
        float(test_coupled_problem.heat_problem.dt),
        float(test_coupled_problem.hydrogen_problem.dt),
    )


def test_dts_always_the_same():
    """Test that the correct dt value is modified"""

    dt = F.Stepsize(
        1000, growth_factor=1000, cutback_factor=0.9, target_nb_iterations=4
    )

    test_heat_problem = F.HeatTransferProblem(
        mesh=test_mesh,
        subdomains=test_subdomains,
        settings=F.Settings(
            atol=1, rtol=0.9999, transient=True, stepsize=dt, final_time=1
        ),
        initial_condition=F.InitialTemperature(value=500, volume=test_subdomains[0]),
    )

    test_hydrogen_problem = F.HydrogenTransportProblem(
        mesh=test_mesh,
        subdomains=test_subdomains,
        species=[test_H],
        initial_conditions=[
            F.InitialConcentration(value=10, species=test_H, volume=test_subdomains[0])
        ],
        settings=F.Settings(
            atol=1, rtol=0.9999, transient=True, stepsize=0.1, final_time=1
        ),
    )

    test_coupled_problem = F.CoupledTransientHeatTransferHydrogenTransport(
        heat_problem=test_heat_problem,
        hydrogen_problem=test_hydrogen_problem,
    )

    test_coupled_problem.initialise()

    value_dt_heat_initial = test_heat_problem.dt.value

    test_coupled_problem.run()

    assert np.isclose(value_dt_heat_initial, 0.1)


def test_error_raised_when_final_times_not_the_same():
    """Test that an error is raised when the final time values given in the heat_problem
    and the hydrogen_problem are not the same"""

    test_heat_problem = F.HeatTransferProblem(
        mesh=test_mesh,
        subdomains=test_subdomains,
        settings=F.Settings(
            atol=1, rtol=0.9999, transient=True, stepsize=1, final_time=10
        ),
    )

    test_hydrogen_problem = F.HydrogenTransportProblem(
        mesh=test_mesh,
        subdomains=test_subdomains,
        species=[test_H],
        settings=F.Settings(
            atol=1, rtol=0.9999, transient=True, stepsize=1, final_time=5
        ),
    )

    test_coupled_problem = F.CoupledTransientHeatTransferHydrogenTransport(
        heat_problem=test_heat_problem,
        hydrogen_problem=test_hydrogen_problem,
    )

    with pytest.raises(
        ValueError,
        match="Final time values in the heat transfer and hydrogen transport model"
        " must be the same",
    ):
        test_coupled_problem.initialise()


def test_error_raised_when_both_problems_not_transient():
    """Test that an error is raised when the heat_problem and hydorgen_problem are not
    set to transient"""

    test_heat_problem = F.HeatTransferProblem(
        settings=F.Settings(atol=1, rtol=0.9999, transient=False),
    )

    test_hydrogen_problem = F.HydrogenTransportProblem(
        settings=F.Settings(
            atol=1, rtol=0.9999, transient=True, stepsize=1, final_time=4
        ),
    )

    with pytest.raises(
        TypeError,
        match="Both the heat and hydrogen problems must be set to transient",
    ):
        F.CoupledTransientHeatTransferHydrogenTransport(
            heat_problem=test_heat_problem,
            hydrogen_problem=test_hydrogen_problem,
        )



from collections.abc import Callable

import dolfinx
import numpy as np
import ufl
from dolfinx import fem
from packaging import version


def as_fenics_constant(
    value: float | int | fem.Constant, mesh: dolfinx.mesh.Mesh
) -> fem.Constant:
    """Converts a value to a dolfinx.Constant

    Args:
        value: the value to convert
        mesh: the mesh of the domiain

    Returns:
        The converted value

    Raises:
        TypeError: if the value is not a float, an int or a dolfinx.Constant
    """
    if isinstance(value, float | int):
        return fem.Constant(mesh, dolfinx.default_scalar_type(float(value)))
    elif isinstance(value, fem.Constant):
        return value
    else:
        raise TypeError(
            f"Value must be a float, an int or a dolfinx.Constant, not {type(value)}"
        )


def as_mapped_function(
    value: Callable,
    function_space: fem.FunctionSpace | None = None,
    t: fem.Constant | None = None,
    temperature: fem.Function | fem.Constant | ufl.core.expr.Expr | None = None,
) -> ufl.core.expr.Expr:
    """Maps a user given callable function to the mesh, time or temperature within
    festim as needed

    Args:
        value: the callable to convert
        function_space: the function space of the domain, optional
        t: the time, optional
        temperature: the temperature, optional

    Returns:
        The mapped function
    """

    # Extract the input variable names in the callable function `value`
    arguments = value.__code__.co_varnames

    kwargs = {}
    if "t" in arguments:
        kwargs["t"] = t
    if "x" in arguments:
        x = ufl.SpatialCoordinate(function_space.mesh)
        kwargs["x"] = x
    if "T" in arguments:
        kwargs["T"] = temperature

    return value(**kwargs)


def as_fenics_interp_expr_and_function(
    value: Callable,
    function_space: dolfinx.fem.function.FunctionSpace,
    t: fem.Constant | None = None,
    temperature: fem.Function | fem.Constant | ufl.core.expr.Expr | None = None,
) -> tuple[fem.Expression, fem.Function]:
    """Takes a user given callable function, maps the function to the mesh, time or
    temperature within festim as needed. Then creates the fenics interpolation
    expression and function objects

    Args:
        value: the callable to convert
        function_space: The function space to interpolate function over
        t: the time, optional
        temperature: the temperature, optional

    Returns:
        fenics interpolation expression, fenics function
    """

    mapped_function = as_mapped_function(
        value=value, function_space=function_space, t=t, temperature=temperature
    )

    fenics_interpolation_expression = fem.Expression(
        mapped_function,
        get_interpolation_points(function_space.element),
    )

    fenics_object = fem.Function(function_space)
    fenics_object.interpolate(fenics_interpolation_expression)

    return fenics_interpolation_expression, fenics_object


class Value:
    """
    A class to handle input values from users and convert them to a relevent fenics
    object

    Args:
        input_value: The value of the user input

    Attributes:
        input_value : The value of the user input
        fenics_interpolation_expression : The expression of the user input that is used
            to update the `fenics_object`
        fenics_object : The value of the user input in fenics format
        explicit_time_dependent : True if the user input value is explicitly time
            dependent
        temperature_dependent : True if the user input value is temperature dependent

    """

    input_value: (
        float
        | int
        | fem.Constant
        | np.ndarray
        | fem.Expression
        | ufl.core.expr.Expr
        | fem.Function
    )

    ufl_expression: ufl.core.expr.Expr
    fenics_interpolation_expression: fem.Expression
    fenics_object: fem.Function | fem.Constant | ufl.core.expr.Expr
    explicit_time_dependent: bool
    temperature_dependent: bool

    def __init__(self, input_value):
        self.input_value = input_value

        self.ufl_expression = None
        self.fenics_interpolation_expression = None
        self.fenics_object = None

    def __repr__(self) -> str:
        return str(self.input_value)

    @property
    def input_value(self):
        return self._input_value

    @input_value.setter
    def input_value(self, value):
        if value is None:
            self._input_value = value
        elif isinstance(
            value,
            float
            | int
            | fem.Constant
            | np.ndarray
            | fem.Expression
            | ufl.core.expr.Expr
            | fem.Function,
        ):
            self._input_value = value
        elif callable(value):
            self._input_value = value
        else:
            raise TypeError(
                "Value must be a float, int, fem.Constant, np.ndarray, fem.Expression,"
                f" ufl.core.expr.Expr, fem.Function, or callable not {value}"
            )

    @property
    def explicit_time_dependent(self) -> bool:
        """Returns true if the value given is time dependent"""
        if self.input_value is None:
            return False
        if isinstance(self.input_value, fem.Constant | ufl.core.expr.Expr):
            return False
        if callable(self.input_value):
            arguments = self.input_value.__code__.co_varnames
            return "t" in arguments
        else:
            return False

    @property
    def temperature_dependent(self) -> bool:
        """Returns true if the value given is temperature dependent"""
        if self.input_value is None:
            return False
        if isinstance(self.input_value, fem.Constant | ufl.core.expr.Expr):
            return False
        if callable(self.input_value):
            arguments = self.input_value.__code__.co_varnames
            return "T" in arguments
        else:
            return False

    def convert_input_value(
        self,
        function_space: dolfinx.fem.function.FunctionSpace | None = None,
        t: fem.Constant | None = None,
        temperature: fem.Function | fem.Constant | ufl.core.expr.Expr | None = None,
        up_to_ufl_expr: bool | None = False,
    ):
        """Converts a user given value to a relevent fenics object depending
        on the type of the value provided

        Args:
            function_space: the function space of the fenics object, optional
            t: the time, optional
            temperature: the temperature, optional
            up_to_ufl_expr: if True, the value is only mapped to a function if the input
                is callable, not interpolated or converted to a function, optional
        """
        if isinstance(
            self.input_value, fem.Constant | fem.Function | ufl.core.expr.Expr
        ):
            self.fenics_object = self.input_value

        elif isinstance(self.input_value, fem.Expression):
            self.fenics_interpolation_expression = self.input_value

        elif isinstance(self.input_value, float | int):
            self.fenics_object = as_fenics_constant(
                value=self.input_value, mesh=function_space.mesh
            )

        elif callable(self.input_value):
            args = self.input_value.__code__.co_varnames
            # if only t is an argument, create constant object
            if "t" in args and "x" not in args and "T" not in args:
                if not isinstance(self.input_value(t=float(t)), float | int):
                    raise ValueError(
                        "self.value should return a float or an int, not "
                        + f"{type(self.input_value(t=float(t)))} "
                    )

                self.fenics_object = as_fenics_constant(
                    value=self.input_value(t=float(t)), mesh=function_space.mesh
                )

            elif up_to_ufl_expr:
                self.fenics_object = as_mapped_function(
                    value=self.input_value,
                    function_space=function_space,
                    t=t,
                    temperature=temperature,
                )

            else:
                self.fenics_interpolation_expression, self.fenics_object = (
                    as_fenics_interp_expr_and_function(
                        value=self.input_value,
                        function_space=function_space,
                        t=t,
                        temperature=temperature,
                    )
                )

    def update(self, t: float):
        """Updates the value

        Args:
            t: the time
        """
        if callable(self.input_value):
            arguments = self.input_value.__code__.co_varnames

            if isinstance(self.fenics_object, fem.Constant) and "t" in arguments:
                self.fenics_object.value = float(self.input_value(t=t))

            elif isinstance(self.fenics_object, fem.Function):
                if self.fenics_interpolation_expression is not None:
                    self.fenics_object.interpolate(self.fenics_interpolation_expression)


# Check the version of dolfinx
dolfinx_version = dolfinx.__version__

# Define the appropriate method based on the version
if version.parse(dolfinx_version) > version.parse("0.9.0"):
    get_interpolation_points = lambda element: element.interpolation_points
else:
    get_interpolation_points = lambda element: element.interpolation_points()


def nmm_interpolate(
    f_out: fem.Function,
    f_in: fem.Function,
    cells: dolfinx.mesh.MeshTags | None = None,
    padding: float | None = 1e-11,
):
    """Non Matching Mesh Interpolate: interpolate one function (f_in) from one mesh into
    another function (f_out) with a mismatching mesh

    args:
        f_out: function to interpolate into
        f_in: function to interpolate from

    notes:
    https://fenicsproject.discourse.group/t/gjk-error-in-interpolation-between-non-matching-second-ordered-3d-meshes/16086/6
    """

    if cells is None:
        dim = f_out.function_space.mesh.topology.dim
        index_map = f_out.function_space.mesh.topology.index_map(dim)
        ncells = index_map.size_local + index_map.num_ghosts
        cells = np.arange(ncells, dtype=np.int32)

    interpolation_data = fem.create_interpolation_data(
        f_out.function_space, f_in.function_space, cells, padding=padding
    )
    f_out.interpolate_nonmatching(f_in, cells, interpolation_data=interpolation_data)


def is_it_time_to_export(
    times: list | None, current_time: float, atol=0, rtol=1.0e-5
) -> bool:
    """
    Checks if the exported field should be written to a file or not based on the
    current time and the times in `export.times`

    Args:
        current_time: the current simulation time
        atol: absolute tolerance for time comparison
        rtol: relative tolerance for time comparison
        times: the times at which the field should be exported, if None, returns True

    Returns:
        bool: True if the exported field should be written to a file, else False
    """
    if times is None:
        return True

    for time in times:
        if np.isclose(time, current_time, atol=atol, rtol=rtol):
            return True

    return False



import numpy as np

import festim as F

my_model = F.HydrogenTransportProblemDiscontinuous()


N = 1500
vertices = np.linspace(0, 1, num=N)
my_model.mesh = F.Mesh1D(vertices)

material = F.Material(D_0=2.0, E_D=0)

material.K_S_0 = 2.0
material.E_K_S = 0

subdomain = F.VolumeSubdomain1D(
    1, borders=[vertices[0], vertices[-1]], material=material
)

left_surface = F.SurfaceSubdomain1D(id=1, x=vertices[0])
right_surface = F.SurfaceSubdomain1D(id=2, x=vertices[-1])

my_model.subdomains = [subdomain, left_surface, right_surface]
my_model.surface_to_volume = {right_surface: subdomain, left_surface: subdomain}

H = F.Species("H", mobile=True)
trapped_H = F.Species("H_trapped", mobile=False)
empty_trap = F.ImplicitSpecies(n=0.5, others=[trapped_H])

my_model.species = [H, trapped_H]

for species in my_model.species:
    species.subdomains = my_model.volume_subdomains


my_model.reactions = [
    F.Reaction(
        reactant=[H, empty_trap],
        product=[trapped_H],
        k_0=2,
        E_k=0,
        p_0=0.1,
        E_p=0,
        volume=domain,
    )
    for domain in my_model.volume_subdomains
]

my_model.boundary_conditions = [
    F.DirichletBC(left_surface, value=0.05, species=H),
    F.DirichletBC(right_surface, value=0.2, species=H),
]


my_model.temperature = lambda x: 300 + 100 * x[0]

my_model.settings = F.Settings(atol=None, rtol=1e-5, transient=False)

my_model.exports = [
    F.VTXSpeciesExport(filename=f"u_{subdomain.id}.bp", field=H, subdomain=subdomain)
    for subdomain in my_model.volume_subdomains
]

my_model.initialise()
my_model.run()



import numpy as np
import ufl

import festim as F

from .tools import error_L2


def test_MMS_coupled_problem():
    """MMS coupled heat and hydrogen test with 1 mobile species and 1 trap in a 1s
    transient, the values of the temperature, mobile and trapped solutions at the last
    time step is compared to an analytical solution"""

    # coupled simulation properties
    density, heat_capacity = 1.2, 2.6
    thermal_conductivity = 4.2
    D_0, E_D = 1.2, 0.1
    k_0, E_k = 2.2, 0.2
    p_0, E_p = 0.5, 0.1
    n_trap = 5
    k_B = F.k_B
    final_time = 1

    # common festim objects
    test_mesh = F.Mesh1D(vertices=np.linspace(0, 1, 2000))
    test_mat = F.Material(
        D_0=D_0,
        E_D=E_D,
        thermal_conductivity=thermal_conductivity,
        density=density,
        heat_capacity=heat_capacity,
    )
    test_vol_sub = F.VolumeSubdomain1D(id=1, borders=[0, 1], material=test_mat)
    left_sub = F.SurfaceSubdomain1D(id=2, x=0)
    right_sub = F.SurfaceSubdomain1D(id=3, x=1)
    test_mobile = F.Species("mobile", mobile=True)
    test_trapped = F.Species(name="trapped", mobile=False, subdomains=[test_vol_sub])
    test_traps = F.ImplicitSpecies(n=n_trap, others=[test_mobile, test_trapped])

    # define temperature sim
    exact_T_solution = lambda x, t: 3 * x[0] ** 2 + 10 * t

    dTdt = 10
    mms_T_source = (
        test_mat.density * test_mat.heat_capacity * dTdt
        - test_mat.thermal_conductivity * 6
    )

    test_heat_problem = F.HeatTransferProblem(
        mesh=test_mesh,
        subdomains=[test_vol_sub, left_sub, right_sub],
        boundary_conditions=[
            F.FixedTemperatureBC(subdomain=left_sub, value=exact_T_solution),
            F.FixedTemperatureBC(subdomain=right_sub, value=exact_T_solution),
        ],
        sources=[F.HeatSource(value=mms_T_source, volume=test_vol_sub)],
        initial_condition=F.InitialTemperature(
            value=lambda x: exact_T_solution(x, 0), volume=test_vol_sub
        ),
        settings=F.Settings(
            atol=1e-10,
            rtol=1e-10,
            transient=True,
            stepsize=final_time / 20,
            final_time=final_time,
        ),
    )

    # define hydrogen problem
    exact_mobile_solution = lambda x, t: 2 * x[0] ** 2 + 15 * t
    exact_trapped_solution = lambda x, t: 4 * x[0] ** 2 + 12 * t

    exact_mobile_intial_cond = lambda x: exact_mobile_solution(x, t=0)
    exact_trapped_intial_cond = lambda x: exact_trapped_solution(x, t=0)

    dmobiledt = 15
    dtrappeddt = 12

    D = lambda x, t: D_0 * ufl.exp(-E_D / (k_B * exact_T_solution(x, t)))
    k = lambda x, t: k_0 * ufl.exp(-E_k / (k_B * exact_T_solution(x, t)))
    p = lambda x, t: p_0 * ufl.exp(-E_p / (k_B * exact_T_solution(x, t)))

    def mms_mobile_source(x, t):
        return (
            dmobiledt
            - ufl.div(D(x, t) * ufl.grad(exact_mobile_solution(x, t)))
            + k(x, t)
            * (exact_mobile_solution(x, t))
            * (n_trap - (exact_trapped_solution(x, t)))
            - p(x, t) * (exact_trapped_solution(x, t))
        )

    def mms_trapped_source(x, t):
        return (
            dtrappeddt
            + k(x, t)
            * (exact_mobile_solution(x, t))
            * (n_trap - (exact_trapped_solution(x, t)))
            - p(x, t) * (exact_trapped_solution(x, t))
        )

    test_hydrogen_problem = F.HydrogenTransportProblem(
        mesh=test_mesh,
        subdomains=[test_vol_sub, left_sub, right_sub],
        boundary_conditions=[
            F.FixedConcentrationBC(
                subdomain=left_sub,
                value=exact_mobile_solution,
                species=test_mobile,
            ),
            F.FixedConcentrationBC(
                subdomain=right_sub,
                value=exact_mobile_solution,
                species=test_mobile,
            ),
            F.FixedConcentrationBC(
                subdomain=left_sub,
                value=exact_trapped_solution,
                species=test_trapped,
            ),
            F.FixedConcentrationBC(
                subdomain=right_sub,
                value=exact_trapped_solution,
                species=test_trapped,
            ),
        ],
        species=[test_mobile, test_trapped],
        reactions=[
            F.Reaction(
                reactant=[test_mobile, test_traps],
                product=test_trapped,
                k_0=k_0,
                E_k=E_k,
                p_0=p_0,
                E_p=E_p,
                volume=test_vol_sub,
            )
        ],
        initial_conditions=[
            F.InitialConcentration(
                value=exact_mobile_intial_cond, species=test_mobile, volume=test_vol_sub
            ),
            F.InitialConcentration(
                value=exact_trapped_intial_cond,
                species=test_trapped,
                volume=test_vol_sub,
            ),
        ],
        sources=[
            F.ParticleSource(
                value=mms_mobile_source, volume=test_vol_sub, species=test_mobile
            ),
            F.ParticleSource(
                value=mms_trapped_source, volume=test_vol_sub, species=test_trapped
            ),
        ],
        settings=F.Settings(
            atol=1e-10,
            rtol=1e-10,
            transient=True,
            stepsize=final_time / 20,
            final_time=final_time,
        ),
    )

    # define coupled problem
    test_coupled_problem = F.CoupledTransientHeatTransferHydrogenTransport(
        heat_problem=test_heat_problem,
        hydrogen_problem=test_hydrogen_problem,
    )
    test_coupled_problem.initialise()
    test_coupled_problem.run()

    # compare computed values with exact solutions
    T_computed = test_coupled_problem.hydrogen_problem.temperature_fenics
    mobile_computed = test_mobile.post_processing_solution
    trapped_computed = test_trapped.post_processing_solution

    exact_final_T = lambda x: exact_T_solution(x, t=final_time)
    exact_final_mobile = lambda x: exact_mobile_solution(x, t=final_time)
    exact_final_trapped = lambda x: exact_trapped_solution(x, t=final_time)

    L2_error_T = error_L2(T_computed, exact_final_T)
    L2_error_mobile = error_L2(mobile_computed, exact_final_mobile)
    L2_error_trapped = error_L2(trapped_computed, exact_final_trapped)

    # TEST ensure L2 error below 2e-7
    assert L2_error_T < 2e-07
    assert L2_error_mobile < 2e-07
    assert L2_error_trapped < 2e-07


def test_coupled_problem_non_matching_mesh():
    """MMS coupled heat and hydrogen test with 1 mobile species in a 1s transient with
    mismatched meshes, the value of the mobile solution at the last time step is
    compared to an analytical solution"""

    # coupled simulation properties
    density, heat_capacity = 1.3, 2.3
    thermal_conductivity = 4.1
    D_0, E_D = 1.3, 0.2
    k_B = F.k_B
    final_time = 5

    # common festim objects
    test_mat = F.Material(
        D_0=D_0,
        E_D=E_D,
        thermal_conductivity=thermal_conductivity,
        density=density,
        heat_capacity=heat_capacity,
    )
    test_vol_sub = F.VolumeSubdomain1D(id=1, borders=[0, 1], material=test_mat)
    left_sub = F.SurfaceSubdomain1D(id=2, x=0)
    right_sub = F.SurfaceSubdomain1D(id=3, x=1)
    test_mobile = F.Species("mobile", mobile=True)

    # define temperature sim
    exact_T_solution = lambda x, t: 2 * x[0] ** 2 + 5 * t

    dTdt = 5

    mms_T_source = (
        test_mat.density * test_mat.heat_capacity * dTdt
        - test_mat.thermal_conductivity * 4
    )

    # Corse mesh for heat trasnfer problem
    test_mesh = F.Mesh1D(vertices=np.linspace(0, 1, 500))

    test_heat_problem = F.HeatTransferProblem(
        mesh=test_mesh,
        subdomains=[test_vol_sub, left_sub, right_sub],
        boundary_conditions=[
            F.FixedTemperatureBC(subdomain=left_sub, value=exact_T_solution),
            F.FixedTemperatureBC(subdomain=right_sub, value=exact_T_solution),
        ],
        sources=[F.HeatSource(value=mms_T_source, volume=test_vol_sub)],
        initial_condition=F.InitialTemperature(
            lambda x: exact_T_solution(x, 0), volume=test_vol_sub
        ),
        settings=F.Settings(
            atol=1e-10,
            rtol=1e-10,
            transient=True,
            final_time=final_time,
            stepsize=final_time / 20,
        ),
    )

    # define hydrogen problem
    exact_mobile_solution = lambda x, t: 4 * x[0] ** 2 + 10 * t

    dmobiledt = 10

    D = lambda x, t: D_0 * ufl.exp(-E_D / (k_B * exact_T_solution(x, t)))

    def mms_mobile_source(x, t):
        return dmobiledt - ufl.div(D(x, t) * ufl.grad(exact_mobile_solution(x, t)))

    # Fine mesh for hydrogen transport problem
    test_mesh_2 = F.Mesh1D(vertices=np.linspace(0, 1, 2000))

    test_hydrogen_problem = F.HydrogenTransportProblem(
        mesh=test_mesh_2,
        subdomains=[test_vol_sub, left_sub, right_sub],
        boundary_conditions=[
            F.FixedConcentrationBC(
                subdomain=left_sub, value=exact_mobile_solution, species=test_mobile
            ),
            F.FixedConcentrationBC(
                subdomain=right_sub, value=exact_mobile_solution, species=test_mobile
            ),
        ],
        species=[test_mobile],
        initial_conditions=[
            F.InitialConcentration(
                value=lambda x: exact_mobile_solution(x, t=0),
                species=test_mobile,
                volume=test_vol_sub,
            ),
        ],
        sources=[
            F.ParticleSource(
                value=mms_mobile_source, volume=test_vol_sub, species=test_mobile
            ),
        ],
        settings=F.Settings(
            atol=1e-10,
            rtol=1e-10,
            transient=True,
            final_time=final_time,
            stepsize=final_time / 20,
        ),
    )

    test_coupled_problem = F.CoupledTransientHeatTransferHydrogenTransport(
        heat_problem=test_heat_problem,
        hydrogen_problem=test_hydrogen_problem,
    )

    test_coupled_problem.initialise()
    test_coupled_problem.run()

    exact_solution = lambda x: exact_mobile_solution(x, t=final_time)

    computed_solution = test_mobile.post_processing_solution

    L2_error = error_L2(computed_solution, exact_solution)

    assert L2_error < 2e-07
